from __future__ import annotations

from spdm.core.geo_object import GeoObject

from ..ontology import ic_antennas 


class ICAntennas(ic_antennas._T_ic_antennas):
    def __view__(self, view="RZ", **kwargs) -> GeoObject:

        geo = {}
        styles = {}
        if view != "RZ":
            geo["antenna"] = [antenna.name for antenna in self.antenna]
            styles["antenna"] = {"$matplotlib": {"color": 'blue'}, "text": True}

        return geo 



from functools import cached_property

from spdm.core.actor import Actor
from spdm.core.sp_property import sp_tree
from .modules.pulse_schedule import PulseSchedule
from .modules.transport_solver_numerics import TransportSolverNumerics

from .utils.logger import logger
from .tokamak import Tokamak


@sp_tree
class Scenario(Actor):
    """
    Scenario

    """

    tokamak: Tokamak

    pulse_schedule: PulseSchedule



import os
import getpass
import datetime

try:
    from ..__version__ import version
except Exception:
    FY_VERSION = "alpha"
else:
    FY_VERSION = version

try:
    from ..extension import tags as extension_tags
except ImportError:
    FY_EXT_VERSION = "n/a"
else:
    FY_EXT_VERSION = extension_tags

FY_LABEL = os.environ.get("FY_LABEL", "fytok")

FY_DEBUG = os.environ.get("FY_DEBUG", False)

FY_VERBOSE = os.environ.get("FY_VERBOSE", "info")

FY_JOBID = f"fytok_{getpass.getuser().lower()}_{os.uname().nodename.lower()}_{os.getpid()}"



for k, v in os.environ.items():
    if k.startswith("FY_"):
        os.environ[f"SP_{k[3:]}"] = v

from spdm.utils import envs as sp_envs


FY_LOGO = rf"""
###################################################################################################

    ______      _____     _
   / ____/_  __|_   _|__ | | __
  / /_  / / / /  | |/ _ \| |/ /
 / __/ / /_/ /   | | (_) |   <
/_/    \__, /    |_|\___/|_|\_\
      /____/

 Copyright (c) 2021-present Zhi YU @ Institute of Plasma Physics Chinese Academy of Sciences 
            
 version = {FY_VERSION}  (spdm={sp_envs.SP_VERSION} {FY_EXT_VERSION})

 Run by {getpass.getuser()} at {datetime.datetime.now().isoformat()}.
 Job ID: {FY_JOBID}

###################################################################################################
"""

__all__ = ["FY_DEBUG", "FY_JOBID", "FY_LOGO", "FY_VERBOSE", "FY_VERSION", "FY_EXT_VERSION"]



import scipy.constants
from spdm.core.expression import zero
from spdm.core.sp_property import sp_tree
from spdm.utils.tags import _not_found_
from fytok.modules.amns_data import amns
from fytok.modules.core_sources import CoreSources
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.utilities import *

PI = scipy.constants.pi


@sp_tree
class Radiation(CoreSources.Source):
    identifier = "radiation"

    code = {
        "name": "radiation",
        "description": """
    Source from   bremsstrahlung and impurity line radiation, and synchrotron radiation 
    Reference:
        Synchrotron radiation
            - Trubnikov, JETP Lett. 16 (1972) 25.
    """,
    }  # type: ignore

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D) -> CoreSources.Source.TimeSlice:
        current: CoreSources.Source.TimeSlice = super().fetch(profiles_1d)

        source_1d = current.profiles_1d

        ne = profiles_1d.electrons.density
        Te = profiles_1d.electrons.temperature

        if ne is _not_found_ or Te is _not_found_:
            raise RuntimeError(f"{ne} {Te}")

        Qrad = sum(
            [
                ne * ion.density * amns[ion.label].radiation(Te)
                for ion in profiles_1d.ion
                if ion.density is not _not_found_
            ],
            zero,
        )

        source_1d.electrons.energy -= Qrad

        return current


CoreSources.Source.register(["radiation"], Radiation)



import pathlib

import numpy as np
from spdm.core.file import File
from spdm.core.entry import Entry
from spdm.utils.logger import logger


def sp_read_iterdb_txt(path: str | pathlib.Path):
    """
    :param file: input file / file path
    :return: profile object
    """
    data = []
    with open(path, mode="r") as fid:
        fid.readline()
        key = None
        value = ""
        for line in fid.readlines():
            if not line.startswith("*"):
                value += line
                continue
            elif key is not None:
                d = []
                for s in value.split():
                    if s.isdigit():
                        d.append(int(s))
                    else:
                        try:
                            t = float(s)
                        except ValueError:
                            d.append(s)
                        else:
                            d.append(t)

                data.append([key] + d)

            key = line[1:].strip()
            value = ""

        # data.append((key,  [float(s) for s in value.split()]))
    return data


def sp_to_imas(data: dict):
    entry = Entry({})
    n = 0
    # fmt:off
    entry["shot"]                                                                                  = data[0][1]
    nx                                                                                        = data[n:=n+1][1]
    nion                                                                                      = data[n:=n+1][1]
    nprim                                                                                     = data[n:=n+1][1]
    nimp                                                                                      = data[n:=n+1][1]
    nneu                                                                                      = data[n:=n+1][1]
    nion_beam                                                                                 = data[n:=n+1][1]
    namep                                                                                     = data[n:=n+1][1:]
    namei                                                                                     = data[n:=n+1][1:]
    namen                                                                                     = data[n:=n+1][1:]
    entry["time"]                                                                                  = data[n:=n+1][1]
    entry["geometric_center/r"]                                                                    = data[n:=n+1][1]   # Rgeom : major radius of geometric center at elevation of magnetic axis, meters
    entry["magnetic_axis/r"]                                                                       = data[n:=n+1][1]   # major radius of magnetic axis, meters
    r0                                                                                        = data[n:=n+1][1]   # major radius of vacuum btor ref location, meters
    entry["equilibrium/vacuum_toroidal_field/r0"]                                                  = r0 
    entry["equilibrium/time_slice/0/vacuum_toroidal_field/r0"]                                     = r0 
    entry["core_profiles/time_slice/0/vacuum_toroidal_field/r0"]                                   = r0 
    entry["equilibrium/time_slice/0/boundary/elongation"]                                          = data[n:=n+1][1]   #  kappa : plasma elongation
    entry["equilibrium/time_slice/0/boundary/triangularity"]                                       = data[n:=n+1][1]   #  delta : plasma triangularity
    entry["equilibrium/time_slice/0/boundary/indentation"]                                         = data[n:=n+1][1]   #  pindent : plasma indentation
    entry["equilibrium/time_slice/0/global_quantities/volume"]                                     = data[n:=n+1][1]   #  volo : plasma volume, meters**3
    entry["equilibrium/time_slice/0/global_quantities/area"]                                       = data[n:=n+1][1]   #  cxareao : plasma cross-sectional area, meters**2
    b0                                                                                        = data[n:=n+1][1]  #  Btor : vacuum toroidal field at rmajor, tesla
    entry["equilibrium/vacuum_toroidal_field/b0"]                                                  = np.asarray([b0])
    entry["equilibrium/time_slice/0/vacuum_toroidal_field/b0"]                                     = b0
    entry["core_profiles/time_slice/0/vacuum_toroidal_field/b0"]                                   = b0
    entry["core_profiles/time_slice/0/global_quantities/current"]                                  = data[n:=n+1][1]   #  total, ohmic, bootstrap, beam and RF currents, amps
    entry["core_profiles/time_slice/0/global_quantities/current_ohmic"]                            = data[n][2]        #  ohmic current
    entry["core_profiles/time_slice/0/global_quantities/current_bootstrap"]                        = data[n][3]        #  bootstrap current
    entry["core_profiles/time_slice/0/global_quantities/current_non_inductive"]                    = data[n][4] + data[n][5] #  beam +RF current
    entry["core_profiles/time_slice/0/global_quantities/beta_pol"]                                 = data[n:=n+1][1]   #  betap : poloidal beta
    entry["core_profiles/time_slice/0/global_quantities/beta_tor"]                                 = data[n:=n+1][1]   #  beta : toroidal beta
    entry["core_profiles/time_slice/0/global_quantities/li_3"]                                     = data[n:=n+1][1]   #  ali : plasma inductance
    entry["core_profiles/time_slice/0/global_quantities/t_e_peaking"]                              = data[n:=n+1][1]   #  te0 : central electron temperature
    entry["core_profiles/time_slice/0/global_quantities/t_i_average_peaking"]                      = data[n:=n+1][1]   #  ti0 : central ion temperature
    psi                                                                                   = np.asarray(data[n:=n+1][1:])  #  psi on rho grid, volt*second/radian
    rho_tor                                                                               = np.asarray(data[n:=n+1][1:])  #  rho grid, meters
    psi_axis                                                                                  = psi[0]
    psi_boundary                                                                              = psi[-1]
    rho_tor_boundary                                                                          = rho_tor[-1]
    entry["core_profiles/time_slice/0/profiles_1d/grid/psi"]                                       = psi 
    entry["core_profiles/time_slice/0/profiles_1d/grid/rho_tor"]                                   = rho_tor
    entry["core_profiles/time_slice/0/profiles_1d/grid/psi_magnetic_axis"]                         = psi_axis
    entry["core_profiles/time_slice/0/profiles_1d/grid/psi_boundary"]                              = psi_boundary 
    entry["core_profiles/time_slice/0/profiles_1d/grid/rho_tor_boundary"]                          = rho_tor_boundary
    entry["core_profiles/time_slice/0/profiles_1d/grid/psi_norm"]                                  = (psi -psi_axis)/(psi_boundary-psi_axis)
    entry["core_profiles/time_slice/0/profiles_1d/grid/rho_tor_norm"]                              = rho_tor/rho_tor_boundary
    entry["core_profiles/time_slice/0/profiles_1d/fcap"]                                           = np.asarray(data[n:=n+1][1:])  #  fcap, (i.e., f(psilim)/f(psi))
    entry["core_profiles/time_slice/0/profiles_1d/gcap"]                                           = np.asarray(data[n:=n+1][1:])  #  gcap, (i.e., <(grad rho)**2*(R0/R)**2>)
    entry["core_profiles/time_slice/0/profiles_1d/hcap"]                                           = np.asarray(data[n:=n+1][1:])  #  hcap, (i.e., (dvolume/drho)/(4*pi*pi*R0*rho))
    entry["core_profiles/time_slice/0/profiles_1d/electrons/temperature"]                          = Te = np.asarray(data[n:=n+1][1:])*1000.0  #  electron temperature, eV
    entry["core_profiles/time_slice/0/profiles_1d/t_i_average"]                                    = Ti = np.asarray(data[n:=n+1][1:])*1000.0  #  ion temperatue, eV
    entry["core_profiles/time_slice/0/profiles_1d/q"]                                              = np.asarray(data[n:=n+1][1:])  # q (i.e., safety factor) profile
    entry["core_profiles/time_slice/0/profiles_1d/electrons/density"]                              = np.asarray(data[n:=n+1][1:])  # electron density, #/meter**3


    for i_prim in range(nprim):        
        
        entry[f"core_profiles/time_slice/0/profiles_1d/ion/{i_prim}/@name"]                        = namep[i_prim].capitalize()               # ion name  
        entry[f"core_profiles/time_slice/0/profiles_1d/ion/{i_prim}/density"]                      = np.asarray(data[n:=n+1][1:]) # primary ion density, #/meter**3, species: d       
        entry[f"core_profiles/time_slice/0/profiles_1d/ion/{i_prim}/temperature"]                  = Ti
    for i_prim in range(nprim):
        entry[f"core_sources/source/ionisation/time_slice/0/profiles_1d/ion/{i_prim}/density"]     = np.asarray(data[n:=n+1][1:]) # ionisation
        entry[f"core_sources/source/recombination/time_slice/0/profiles_1d/ion/{i_prim}/density"]  = np.asarray(data[n:=n+1][1:]) # recombination
        entry[f"core_sources/source/equipartition/time_slice/0/profiles_1d/ion/{i_prim}/density"]  = np.asarray(data[n:=n+1][1:]) # collisional_equipartition ,scx : s
        entry[f"core_sources/source/nbi/time_slice/0/profiles_1d/ion/{i_prim}/density"]            = np.asarray(data[n:=n+1][1:]) # nbi ,sbcx : sink due to cx with be
        entry[f"core_sources/source/total/time_slice/0/profiles_1d/ion/{i_prim}/density"]          = np.asarray(data[n:=n+1][1:]) # total source rate,
        entry[f"core_sources/source/custom_1/time_slice/0/profiles_1d/ion/{i_prim}/density"]       = np.asarray(data[n:=n+1][1:]) # dudt : s dot, #/(meter**3*second)


    for i_prim in range(nprim): 
        entry[f"core_profiles/time_slice/0/profiles_1d/ion/{i_prim}/density_fast"]                 = np.asarray(data[n:=n+1][1:]) # fast ion density, #/meter**3, species: d       

    for i_neu in range(nneu):
        entry[f"core_profiles/time_slice/0/profiles_1d/neutral/{i_neu}/density"]                   = np.asarray(data[n:=n+1][1:]) #  neutral density, #/meter**3, species: d 
        entry[f"core_profiles/time_slice/0/profiles_1d/neutral/{i_neu}/density"]                   = np.asarray(data[n:=n+1][1:]) #  neutral density from wall source, #/meter**3, species: d       
        entry[f"core_profiles/time_slice/0/profiles_1d/neutral/{i_neu}/density"]                   = np.asarray(data[n:=n+1][1:]) #  neutral density from volume source, #/meter**3, species: d       
        entry[f"core_sources/source/particles_to_wall/time_slice/0/profiles_1d/neutral/{i_neu}/density"] = np.asarray(data[n:=n+1][1:]) #  volume source of neutrals, #/(meter**3*second), species: d       

    entry["core_sources/source/nbi/time_slice/0/profiles_1d/electrons/particle"]                   = np.asarray(data[n:=n+1][1:]) # beam electron source, #/(meter**3*second)
    entry["core_sources/source/nbi/time_slice/0/profiles_1d/ion/0/particle"]                       = np.asarray(data[n:=n+1][1:]) # beam ion source, #/(meter**3*second)
    entry["core_profiles/time_slice/0/profiles_1d/j_total"]                                        = np.asarray(data[n:=n+1][1:]) #  total current density, amps/meter**2
    entry["core_profiles/time_slice/0/profiles_1d/j_ohmic"]                                        = np.asarray(data[n:=n+1][1:]) # ohmic current density, amps/meter**2
    entry["core_profiles/time_slice/0/profiles_1d/j_bootstrap"]                                    = np.asarray(data[n:=n+1][1:]) # bootstrap current density, amps/meter**2
    entry["core_profiles/time_slice/0/profiles_1d/j_non_inductive"]                                =(np.asarray(data[n:=n+1][1:]) # beam-driven current density, amps/meter**2
                                                                                                    + np.asarray(data[n:=n+1][1:])) # RF current density, amps/meter**2
    entry["core_profiles/time_slice/0/profiles_1d/_rbfgh"]                                         = np.asarray(data[n:=n+1][1:]) # rho*bp0*fcap*gcap*hcap, tesla*meters
    entry["core_profiles/time_slice/0/profiles_1d/zeff"]                                           = np.asarray(data[n:=n+1][1:]) # rho*bp0*fcap*gcap*hcap, tesla*meters
    entry["core_profiles/time_slice/0/profiles_1d/rotation_frequency_tor_sonic"]                   = np.asarray(data[n:=n+1][1:]) # angular rotation speed profile, rad/sec

    entry["core_transport/model/0/time_slice/0/profiles_1d/electrons/particles/d"]                 = np.asarray(data[n:=n+1][1:]) # electron thermal diffusivity, meters**2/sec on half grid
    entry["core_transport/model/0/time_slice/0/profiles_1d/ion/0/particles/d"]                     = np.asarray(data[n:=n+1][1:]) # ion thermal diffusivity, meters**2/second on half grid
    entry["core_transport/model/5/time_slice/0/profiles_1d/ion/0/energy/d"]                        = np.asarray(data[n:=n+1][1:]) # ion neoclassical thermal conductivity, 1/(meter*second), on half grid
    entry["core_transport/model/0/time_slice/0/profiles_1d/electrons/dpdt"]                        = np.asarray(data[n:=n+1][1:]) # wdot, electrons, watts/meter**3
    entry["core_transport/model/0/time_slice/0/profiles_1d/ion/0/dpdt"]                            = np.asarray(data[n:=n+1][1:]) # wdot, ions, watts/meter**3
    entry["core_transport/model/0/time_slice/0/profiles_1d/conductivity_parallel"]                 = np.asarray(data[n:=n+1][1:]) # electron conduction, watts/meter**3
    entry["core_transport/model/0/time_slice/0/profiles_1d/ion/0/conductivity_parallel"]           = np.asarray(data[n:=n+1][1:]) # ion conduction, watts/meter**3
    entry["core_transport/model/0/time_slice/0/profiles_1d/electrons/particles/v"]                 = np.asarray(data[n:=n+1][1:]) # electron convection, watts/meter**3
    entry["core_transport/model/0/time_slice/0/profiles_1d/ion/0/particles/v"]                     = np.asarray(data[n:=n+1][1:]) # ion convection, watts/meter**3

    # Energy Source
    entry["core_sources/source/nbi/time_slice/0/profiles_1d/electrons/energy"]                     = np.asarray(data[n:=n+1][1:]) #  power to elec. from beam, watts/meter**3
    entry["core_sources/source/equipartition/time_slice/0/profiles_1d/electrons/energy"]           = np.asarray(data[n:=n+1][1:]) #  qdelt : electron-ion equilibration, watts/meter**3
    entry["core_sources/source/nbi/time_slice/0/profiles_1d/ion/0/energy"]                         = np.asarray(data[n:=n+1][1:]) #  power to ions from beam, watts/meter**3
    entry["core_sources/source/auxiliary/time_slice/0/profiles_1d/electrons/energy"]               = np.asarray(data[n:=n+1][1:]) #  qrfe, RF electron heating, watts/meter**3
    entry["core_sources/source/auxiliary/time_slice/0/profiles_1d/ion/0/energy"]                   = np.asarray(data[n:=n+1][1:]) #  qrfi, -RF ion heating, watts/meter**3
    entry["core_sources/source/recombination/time_slice/0/profiles_1d/electrons/energy"]           = np.asarray(data[n:=n+1][1:]) #  qione, electron power density due to recombination and impact ionization, watts/meter**3
    entry["core_sources/source/recombination/time_slice/0/profiles_1d/ion/0/energy"]               = np.asarray(data[n:=n+1][1:]) #  qioni, ion power density due to recombination and impact ionization, watts/meter**3
    entry["core_sources/source/charge_exchange/time_slice/0/profiles_1d/ion/0/energy"]             = np.asarray(data[n:=n+1][1:]) #  qcx, ion power density due to neutral-ion charge exchange, watts/meter**3
    entry["core_sources/source/2d/time_slice/0/profiles_1d/electrons/energy"]                      = np.asarray(data[n:=n+1][1:]) #  2d electron heating, watts/meter**3
    entry["core_sources/source/2d/time_slice/0/profiles_1d/ion/0/energy"]                          = np.asarray(data[n:=n+1][1:]) #  2d ion heating, watts/meter**3
    entry["core_sources/source/fusion/time_slice/0/profiles_1d/electrons/energy"]                  = np.asarray(data[n:=n+1][1:]) #  fusion electron heating, watts/meter**3
    entry["core_sources/source/fusion/time_slice/0/profiles_1d/ion/0/energy"]                      = np.asarray(data[n:=n+1][1:]) #  fusion ion heating, watts/meter**3
    entry["core_sources/source/beam/time_slice/0/profiles_1d/electrons/energy"]                    = np.asarray(data[n:=n+1][1:]) #  beam fusion electron heating, watts/meter**3
    entry["core_sources/source/beam/time_slice/0/profiles_1d/ion/0/energy"]                        = np.asarray(data[n:=n+1][1:]) #  beam fusion ion heating, watts/meter**3
    entry["core_sources/source/qmag/time_slice/0/profiles_1d/electrons/energy"]                    = np.asarray(data[n:=n+1][1:]) #  qmag electron heating, watts/meter**3
    entry["core_sources/source/sawtooth/time_slice/0/profiles_1d/electrons/energy"]                = np.asarray(data[n:=n+1][1:]) #  sawtooth electron heating, watts/meter**3
    entry["core_sources/source/sawtooth/time_slice/0/profiles_1d/ion/0/energy"]                    = np.asarray(data[n:=n+1][1:]) #  sawtooth ion  heating, watts/meter**3
    entry["core_sources/source/radiation/time_slice/0/profiles_1d/ion/0/energy"]                   = np.asarray(data[n:=n+1][1:]) #  radiated power density, watts/meter**3
    entry["core_sources/source/ohmic/time_slice/0/profiles_1d/electrons/energy"]                   = np.asarray(data[n:=n+1][1:]) #  (electron) ohmic power density, watts/meter**3

    # flux surface  
    entry["equilibrium/time_slice/0/profiles_1d/major_radius"]                                     = np.asarray(data[n:=n+1][1:]) #  average major radius of each flux surface, meters, evaluated at elevation of magnetic axis
    entry["equilibrium/time_slice/0/profiles_1d/minor_radius"]                                     = np.asarray(data[n:=n+1][1:]) #  average minor radius of each flux surface, meters, evaluated at elevation of magnetic axis
    entry["equilibrium/time_slice/0/profiles_1d/volume"]                                           = np.asarray(data[n:=n+1][1:]) #  volume of each flux surface, meters**3
    entry["equilibrium/time_slice/0/profiles_1d/elongation"]                                       = np.asarray(data[n:=n+1][1:]) #  elongation of each flux surface
    entry["equilibrium/time_slice/0/profiles_1d/triangularity"]                                    = np.asarray(data[n:=n+1][1:]) #  triangularity of each flux surface
    entry["equilibrium/time_slice/0/profiles_1d/indentation"]                                      = np.asarray(data[n:=n+1][1:]) #  indentation of each flux surface
    entry["equilibrium/time_slice/0/profiles_1d/surface"]                                          = np.asarray(data[n:=n+1][1:]) #  surface area of each flux surface, meter**2
    entry["equilibrium/time_slice/0/profiles_1d/Rhr4pp"]                                           = np.asarray(data[n:=n+1][1:]) #  this is 4*pi*pi*R0*hcap*rho*<ABS(grad rho)>; see file "iter1.ps" for explanation
    entry["equilibrium/time_slice/0/profiles_1d/area"]                                             = np.asarray(data[n:=n+1][1:]) #  cross-sectional area of each flux surface, meters**2
    entry["equilibrium/time_slice/0/profiles_1d/gm7"]                                              = np.asarray(data[n:=n+1][1:]) #  flux surface average absolute grad rho
    entry["equilibrium/time_slice/0/profiles_1d/gm3"]                                              = np.asarray(data[n:=n+1][1:]) #  flux surface average (grad rho)**2
        
    nbbdry                                                                                    = data[n:=n+1][1]              #  nplasbdry : number of points on plasma boundary
    entry["equilibrium/time_slice/0/boundary/outline/r"]                                           = np.asarray(data[n:=n+1][1:]) #  r points for plasma boundary, meters
    entry["equilibrium/time_slice/0/boundary/outline/z"]                                           = np.asarray(data[n:=n+1][1:]) #  z points for plasma boundary, meters

    while n < len(data)-1:
        entry[f"core_profiles/time_slice/0/profiles_1d/{data[n:=n+1][0]}"]                         = np.asarray(data[n][1:]) # beam   torque density, nt-m/m**3

    # fmt:on

    return entry


@File.register(["iterdb"])
class ITERDBFile(File):
    """Read iterdb file  (from gacode)"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def read(self) -> Entry:
        if self.url.authority:
            raise NotImplementedError(f"{self.url}")

        path = pathlib.Path(self.url.path)

        if path.suffix.lower() in [".nc", ".h5"]:
            data = File(path, mode="r").read().dump()
        else:
            data = sp_read_iterdb_txt(path)

        return sp_to_imas(data)

    def write(self, d, *args, **kwargs):
        raise NotImplementedError(f"TODO: write ITERDB {self.url}")



from __future__ import annotations


import numpy as np
import scipy.constants

from spdm.core.aos import AoS
from spdm.core.expression import Expression, Variable, zero, derivative
from spdm.core.sp_property import sp_property, sp_tree
from spdm.core.time_series import TimeSeriesAoS
from spdm.core.path import update_tree

from spdm.utils.tags import _not_found_

from ..utils.atoms import atoms
from ..utils.logger import logger

from .equilibrium import Equilibrium
from .utilities import *
from ..ontology import core_profiles, utilities

PI = scipy.constants.pi
TWOPI = 2.0 * PI


@sp_tree
class CoreProfilesSpecies:
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.label is _not_found_ or self.label is None:
            raise RuntimeError(f"Unknown ion ")

        atom_desc = atoms[self.label]

        self._cache["z"] = atom_desc.z
        self._cache["a"] = atom_desc.a

    label: str = sp_property(alias="@name")

    z: float

    a: float

    temperature: Expression = sp_property(units="eV")

    density: Expression = sp_property(units="m^-3")
    # return self.density_thermal + self.density_fast

    density_thermal: Expression = sp_property(units="m^-3")

    density_fast: Expression = sp_property(units="m^-3")

    @sp_property
    def pressure(self) -> Expression:
        # FIXME: coefficient on pressure fast
        return self.density * self.temperature * scipy.constants.electron_volt

    pressure_thermal: Expression = sp_property(units="Pa")

    pressure_fast_perpendicular: Expression = sp_property(units="Pa")

    pressure_fast_parallel: Expression = sp_property(units="Pa")

    rotation_frequency_tor: Expression = sp_property(units="rad.s^-1")

    velocity: CoreVectorComponents = sp_property(units="m.s^-1")


@sp_tree
class CoreProfilesState(CoreProfilesSpecies):
    label: str
    """ String identifying state"""

    electron_configuration: str
    """ Configuration of atomic orbitals of this state, e.g. 1s2-2s1"""

    vibrational_level: float = sp_property(units="Elementary Charge Unit")
    """ Vibrational level (can be bundled)"""

    vibrational_mode: str
    """ Vibrational mode of this state, e.g. _A_g_. Need to define, or adopt a standard
        nomenclature."""

    neutral_type: int
    """ Neutral type (if the considered state is a neutral), in terms of energy. ID =1:
        cold; 2: thermal; 3: fast; 4: NBI"""


@sp_tree
class CoreProfilesIon(CoreProfilesSpecies):
    z_ion_1d: Expression = sp_property(unit="C")

    @sp_property
    def z_ion_square_1d(self) -> Expression:
        return self.z_ion_1d * self.z_ion_1d

    # velocity: _T_core_profiles_vector_components_2 = sp_property(units="m.s^-1")

    @sp_property(unit="s^-1", default_value=0.1)
    def collision_frequency(self) -> Expression:
        r"""
        collision frequency
        $$
            \tau_{ss}^{-1} = \frac{\sqrt{2} \pi e^4 z_s^4 n_{0s}}{m_s^{1/2} T_{0s}^{3/2}} {\rm ln} \Lambda
        $$
        """
        return (
            np.sqrt(2)
            * PI
            * scipy.constants.elementary_charge**4
            * self.z**4
            * self.density_thermal
            / np.sqrt(self.mass)
            / self.temperature**1.5
            / self._parent.coulomb_logarithm
        )


@sp_tree
class CoreProfilesNeutral(CoreProfilesSpecies):
    element: AoS[PlasmaCompositionSpecies]
    """ List of elements forming the atom or molecule"""

    multiple_states_flag: int
    """ Multiple states calculation flag : 0-Only one state is considered; 1-Multiple
        states are considered and are described in the state structure"""

    state: AoS[CoreProfilesState]
    """ Quantities related to the different states of the species (energy, excitation,...)"""


@sp_tree(name="electrons")
class CoreProfilesElectrons(CoreProfilesSpecies):
    label: str = "e"

    @sp_property(units="-")
    def collisionality_norm(self) -> Expression:
        raise NotImplementedError("collisionality_norm")

    @sp_property
    def tau(self):
        return 1.09e16 * ((self.temperature / 1000) ** (3 / 2)) / self.density / self._parent.coulomb_logarithm

    @sp_property
    def vT(self):
        return np.sqrt(self.temperature * scipy.constants.electron_volt / scipy.constants.electron_mass)


@sp_tree(domain="grid")
class CoreProfiles1D(core_profiles._T_core_profiles_profiles_1d):

    grid: CoreRadialGrid = {"extrapolate": 0}

    Electrons = CoreProfilesElectrons
    electrons: CoreProfilesElectrons

    Ion = CoreProfilesIon
    ion: AoS[CoreProfilesIon]

    Neutral = CoreProfilesNeutral
    neutral: AoS[CoreProfilesNeutral]

    rho_tor_norm: array_type | Expression = sp_property(label=r"\bar{\rho}_{tor}", units="-", alias="grid/rho_tor_norm")

    rho_tor: Expression = sp_property(label=r"\rho_{tor}", units="m", alias="grid/rho_tor")

    psi_norm: array_type | Expression = sp_property(label=r"\bar{\psi}", units="-")

    psi: Expression = sp_property(label=r"\psi", units="Wb")

    @sp_property
    def zeff(self) -> Expression:
        return sum([((ion.z_ion_1d**2) * ion.density) for ion in self.ion]) / self.n_i_total

    @sp_property
    def pressure(self) -> Expression:
        return sum([ion.pressure for ion in self.ion], self.electrons.pressure)

    @sp_property
    def pprime(self) -> Expression:
        return self.pressure.d

    @sp_property
    def pressure_thermal(self) -> Expression:
        return sum([ion.pressure_thermal for ion in self.ion], self.electrons.pressure_thermal)

    @sp_property
    def t_i_average(self) -> Expression:
        return sum([ion.z_ion_1d * ion.temperature * ion.density for ion in self.ion]) / self.n_i_total

    @sp_property
    def n_i_total(self) -> Expression:
        return sum([(ion.z_ion_1d * ion.density) for ion in self.ion])

    @sp_property
    def n_i_total_over_n_e(self) -> Expression:
        return self.n_i_total / self.electrons.density

    @sp_property
    def n_i_thermal_total(self) -> Expression:
        return sum([ion.z * ion.density_thermal for ion in self.ion])

    # t_i_average: Expression = sp_property(units="eV")

    # t_i_average_fit: _T_core_profiles_1D_fit = sp_property(units="eV")

    # n_i_total_over_n_e: Expression = sp_property(units="-")

    # n_i_thermal_total: Expression = sp_property(units="m^-3")

    momentum_tor: Expression = sp_property(units="kg.m^-1.s^-1")

    zeff: Expression = sp_property(units="-")

    # zeff_fit: _T_core_profiles_1D_fit = sp_property(units="-")

    pressure_ion_total: Expression = sp_property(units="Pa")

    pressure_thermal: Expression = sp_property(units="Pa")

    pressure_perpendicular: Expression = sp_property(units="Pa")

    pressure_parallel: Expression = sp_property(units="Pa")

    j_total: Expression = sp_property(units="A/m^2")

    @sp_property(units="A")
    def current_parallel_inside(self) -> Expression:
        return self.j_total.I

    j_tor: Expression = sp_property(units="A/m^2")

    j_ohmic: Expression = sp_property(units="A/m^2")

    @sp_property(units="A/m^2")
    def j_non_inductive(self) -> Expression:
        return self.j_total - self.j_ohmic

    j_bootstrap: Expression = sp_property(units="A/m^2")

    @sp_property(units="ohm^-1.m^-1")
    def conductivity_parallel(self) -> Expression:
        return self.j_ohmic / self.e_field.parallel

    @sp_tree
    class EFieldVectorComponents:
        radial: Expression = sp_property(default_value=0.0)

        diamagnetic: Expression

        parallel: Expression

        poloidal: Expression

        toroidal: Expression

        @sp_property
        def parallel(self) -> Expression:
            vloop = self._parent.get("vloop", None)
            if vloop is None:
                logger.error(f"Can not calculate E_parallel from vloop!")
                e_par = 0.0
            else:
                e_par = vloop / (TWOPI * self._parent.grid.r0)
            return e_par

    e_field: EFieldVectorComponents = sp_property(units="V.m^-1")

    phi_potential: Expression = sp_property(units="V")

    rotation_frequency_tor_sonic: Expression

    q: Expression = sp_property(units="-")

    @sp_property(units="-")
    def magnetic_shear(self) -> Expression:
        return self.grid.rho_tor_norm * self.q.dln

    @sp_property
    def beta_pol(self) -> Expression:
        return 4 * self.pressure.I / (self._parent.vacuum_toroidal_field.r0 * constants.mu_0 * (self.j_total**2))

    # if isinstance(d, np.ndarray) or (hasattr(d.__class__, 'empty') and not d.empty):
    #     return d

    # else:
    #     Te = self.electrons.temperature
    #     ne = self.electrons.density

    #     # Electron collisions: Coulomb logarithm
    #     # clog = np.asarray([
    #     #     (24.0 - 1.15*np.log10(ne[idx]*1.0e-6) + 2.30*np.log10(Te[idx]))
    #     #     if Te[idx] >= 10 else (23.0 - 1.15*np.log10(ne[idx]*1.0e-6) + 3.45*np.log10(Te[idx]))
    #     #     for idx in range(len(ne))
    #     # ])
    #     clog = self.coulomb_logarithm
    #     # electron collision time:
    #     # tau_e = (np.sqrt(2.*constants.electron_mass)*(Te**1.5)) / 1.8e-19 / (ne * 1.0e-6) / clog

    #     # Plasma electrical conductivity:
    #     return 1.96e0 * constants.elementary_charge**2   \
    #         * ((np.sqrt(2.*constants.electron_mass)*(Te**1.5)) / 1.8e-19 / clog) \
    #         / constants.m_e

    @sp_property
    def coulomb_logarithm(self) -> Expression:
        """Coulomb logarithm,
        @ref: Tokamaks 2003  Ch.14.5 p727 ,2003
        """
        Te = self.electrons.temperature
        Ne = self.electrons.density

        # Coulomb logarithm
        #  Ch.14.5 p727 Tokamaks 2003

        return (14.9 - 0.5 * np.log(Ne / 1e20) + np.log(Te / 1000)) * (Te < 10) + (
            15.2 - 0.5 * np.log(Ne / 1e20) + np.log(Te / 1000)
        ) * (Te >= 10)

    @sp_property
    def electron_collision_time(self) -> Expression:
        """electron collision time ,
        @ref: Tokamak 2003, eq 14.6.1
        """
        Te = self.electrons.temperature(self.grid.rho_tor_norm)
        Ne = self.electrons.density(self.grid.rho_tor_norm)
        lnCoul = self.coulomb_logarithm(self.grid.rho_tor_norm)
        return 1.09e16 * ((Te / 1000.0) ** (3 / 2)) / Ne / lnCoul

    ffprime: Expression = sp_property(label="$ff^{\prime}$")

    pprime: Expression = sp_property(label="$p^{\prime}$")


@sp_tree
class CoreGlobalQuantities(core_profiles._T_core_profiles_global_quantities):
    vacuum_toroidal_field: VacuumToroidalField

    ip: float = sp_property(units="A")

    current_non_inductive: float = sp_property(units="A")

    current_bootstrap: float = sp_property(units="A")

    v_loop: float = sp_property(units="V")

    li_3: float = sp_property(units="-")

    beta_tor: float = sp_property(units="-")

    beta_tor_norm: float = sp_property(units="-")

    beta_pol: float = sp_property(units="-")

    energy_diamagnetic: float = sp_property(units="J")

    z_eff_resistive: float = sp_property(units="-")

    t_e_peaking: float = sp_property(units="-")

    t_i_average_peaking: float = sp_property(units="-")

    resistive_psi_losses: float = sp_property(units="Wb")

    ejima: float = sp_property(units="-")

    t_e_volume_average: float = sp_property(units="eV")

    n_e_volume_average: float = sp_property(units="m^-3")

    @sp_tree
    class GlobalQuantitiesIon:
        t_i_volume_average: float = sp_property(units="eV")
        n_i_volume_average: float = sp_property(units="m^-3")

    ion: AoS[GlobalQuantitiesIon]

    ion_time_slice: float = sp_property(units="s")


@sp_tree
class CoreProfilesTimeSlice(TimeSlice):

    Profiles1D = CoreProfiles1D

    GlobalQuantities = CoreGlobalQuantities

    profiles_1d: CoreProfiles1D = {}

    global_quantities: CoreGlobalQuantities = {}

    vacuum_toroidal_field: VacuumToroidalField = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # grid: CoreRadialGrid = self.get_cache("profiles_1d/grid", _not_found_)

        # if grid is _not_found_ or Path("psi_axis").get(grid, ...) is ...:
        #     eq_grid: CoreRadialGrid = self._parent.inports["equilibrium/time_slice/0/profiles_1d/grid"].fetch()

        #     if grid is _not_found_:
        #         grid = eq_grid
        #     else:
        #         grid["psi_axis"] = eq_grid.psi_axis
        #         grid["psi_boundary"] = eq_grid.psi_boundary
        #         grid["rho_tor_boundary"] = eq_grid.rho_tor_boundary

        #     self["profiles_1d/grid"] = grid


@sp_tree
class CoreProfiles(IDS):

    ids_properties: IDSProperties

    TimeSlice = CoreProfilesTimeSlice

    time_slice: TimeSeriesAoS[CoreProfilesTimeSlice]

    def preprocess(self, *args, **kwargs) -> CoreProfilesTimeSlice:
        current: CoreProfilesTimeSlice = super().preprocess(*args, **kwargs)

        grid = current.get_cache("profiles_1d/grid", _not_found_)

        if not isinstance(grid, CoreRadialGrid):
            eq_grid: CoreRadialGrid = self.inports["equilibrium/time_slice/0/profiles_1d/grid"].fetch()

            if isinstance(grid, dict):
                new_grid = grid

                if not isinstance(grid.get("psi_axis", _not_found_), float):
                    new_grid["psi_axis"] = eq_grid.psi_axis
                    new_grid["psi_boundary"] = eq_grid.psi_boundary
                    new_grid["rho_tor_boundary"] = eq_grid.rho_tor_boundary
            else:
                rho_tor_norm = kwargs.get("rho_tor_norm", _not_found_)

                if rho_tor_norm is _not_found_:
                    rho_tor_norm = self.code.parameters.rho_tor_norm

                new_grid = eq_grid.remesh(rho_tor_norm)

            current["profiles_1d/grid"] = new_grid

        return current

    def refresh(self, *args, equilibrium: Equilibrium = None, **kwargs):
        return super().refresh(*args, equilibrium=equilibrium, **kwargs)

    def fetch(self, *args, **kwargs) -> CoreProfilesTimeSlice:
        return super().fetch(*args, **kwargs)



from __future__ import annotations

from copy import copy
import math
from spdm.core.aos import AoS
from spdm.core.sp_property import sp_property, sp_tree
from spdm.core.time_series import TimeSeriesAoS
from spdm.core.expression import Expression
from spdm.utils.tags import _not_found_
from .utilities import *
from .core_profiles import CoreProfiles
from .equilibrium import Equilibrium
from ..utils.atoms import atoms
from ..utils.logger import logger

from ..ontology import core_transport


@sp_tree
class CoreTransportModelParticles(core_transport._T_core_transport_model_2_density):
    d: Expression = sp_property(coordinate1=".../grid_d/rho_tor_norm")
    v: Expression = sp_property(coordinate1=".../grid_v/rho_tor_norm")
    flux: Expression = sp_property(coordinate1=".../grid_flux/rho_tor_norm")


@sp_tree
class CoreTransportModelEnergy(core_transport._T_core_transport_model_2_energy):
    d: Expression = sp_property(coordinate1=".../grid_d/rho_tor_norm")
    v: Expression = sp_property(coordinate1=".../grid_v/rho_tor_norm")
    flux: Expression = sp_property(coordinate1=".../grid_flux/rho_tor_norm")


@sp_tree
class CoreTransportModelMomentum(core_transport._T_core_transport_model_4_momentum):
    d: Expression = sp_property(coordinate1=".../grid_d/rho_tor_norm")
    v: Expression = sp_property(coordinate1=".../grid_v/rho_tor_norm")
    flux: Expression = sp_property(coordinate1=".../grid_flux/rho_tor_norm")


@sp_tree
class CoreTransportElectrons(core_transport._T_core_transport_model_electrons):
    label: str = "electrons"
    """ String identifying the neutral species (e.g. H, D, T, He, C, ...)"""

    particles: CoreTransportModelParticles
    energy: CoreTransportModelEnergy
    momentum: CoreTransportModelMomentum


@sp_tree
class CoreTransportIon(core_transport._T_core_transport_model_ions):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        ion = atoms[self.label]
        self.z = ion.z
        self.a = ion.a

    label: str = sp_property(alias="@name")
    """ String identifying the neutral species (e.g. H, D, T, He, C, ...)"""

    z: int
    """ Charge number of the neutral species"""

    a: float
    """ Mass number of the neutral species"""

    particles: CoreTransportModelParticles
    energy: CoreTransportModelEnergy
    momentum: CoreTransportModelMomentum


@sp_tree
class CoreTransportNeutral(core_transport._T_core_transport_model_neutral):
    particles: CoreTransportModelParticles
    energy: CoreTransportModelEnergy


@sp_tree
class CoreTransportProfiles1D(core_transport._T_core_transport_model_profiles_1d):
    grid_d: CoreRadialGrid

    @sp_property
    def grid_v(self) -> CoreRadialGrid:
        return self.grid_d.remesh(self.grid_d.rho_tor_norm)

    @sp_property
    def grid_flux(self) -> CoreRadialGrid:
        rho_tor_norm = self.grid_d.rho_tor_norm
        return self.grid_d.remesh(0.5 * (rho_tor_norm[:-1] + rho_tor_norm[1:]))

    Electrons = CoreTransportElectrons
    electrons: CoreTransportElectrons

    Ion = CoreTransportIon
    ion: AoS[CoreTransportIon] = sp_property(identifier="label", default_initial={})

    Neutral = CoreTransportNeutral
    neutral: AoS[CoreTransportNeutral] = sp_property(identifier="label", default_initial={})


@sp_tree
class CoreTransportTimeSlice(TimeSlice):
    vacuum_toroidal_field: VacuumToroidalField

    flux_multiplier: float = sp_property(default_value=0)

    Profiles1D = CoreTransportProfiles1D
    profiles_1d: CoreTransportProfiles1D


@sp_tree
class CoreTransportModel(Module):
    _plugin_prefix = "fytok.modules.core_transport.model."

    identifier: str

    TimeSlice = CoreTransportTimeSlice
    time_slice: TimeSeriesAoS[CoreTransportTimeSlice]

    def preprocess(self, *args, **kwargs) -> CoreTransportTimeSlice:
        current: CoreTransportTimeSlice = super().preprocess(*args, **kwargs)

        current["vacuum_toroidal_field"] = self.inports["/equilibrium/time_slice/0/vacuum_toroidal_field"].fetch()

        grid = current.get_cache("profiles_1d/grid_d", _not_found_)

        if not isinstance(grid, CoreRadialGrid):
            eq_grid: CoreRadialGrid = self.inports["/equilibrium/time_slice/0/profiles_1d/grid"].fetch()

            if isinstance(grid, dict):
                new_grid = grid
                if not isinstance(grid.get("psi_axis", _not_found_), float):
                    new_grid["psi_axis"] = eq_grid.psi_axis
                    new_grid["psi_boundary"] = eq_grid.psi_boundary
                    new_grid["rho_tor_boundary"] = eq_grid.rho_tor_boundary
                # new_grid = {
                #     **eq_grid._cache,
                #     **{k: v for k, v in grid.items() if v is not _not_found_ and v is not None},
                # }
            else:
                rho_tor_norm = kwargs.get("rho_tor_norm", self.code.parameters.rho_tor_norm)
                new_grid = eq_grid.remesh(rho_tor_norm)

            current["profiles_1d/grid_d"] = new_grid

        return current

    def execute(self, current: CoreTransportTimeSlice, *previous: CoreTransportTimeSlice) -> CoreTransportTimeSlice:
        return super().execute(current, *previous)

    def postprocess(self, current: CoreTransportTimeSlice) -> CoreTransportTimeSlice:
        return super().postprocess(current)

    def fetch(self, *args, **kwargs) -> CoreTransportTimeSlice:
        if len(args) > 0 and isinstance(args[0], CoreProfiles.TimeSlice.Profiles1D):
            args = (args[0].rho_tor_norm, *args[1:])

        return super().fetch(*args, **kwargs)

    def flush(self) -> CoreTransportTimeSlice:
        super().flush()

        current = self.time_slice.current

        profiles_1d: CoreProfiles.TimeSlice.Profiles1D = self.inports["core_profiles/time_slice/0/profiles_1d"].fetch()

        current.update(self.fetch(profiles_1d)._cache)

        return current

    def refresh(
        self, *args, core_profiles: CoreProfiles = None, equilibrium: Equilibrium = None, **kwargs
    ) -> CoreTransportTimeSlice:
        return super().refresh(*args, core_profiles=core_profiles, equilibrium=equilibrium, **kwargs)

    @staticmethod
    def _flux2DV(
        spec: CoreTransport.Model.TimeSlice.Profiles1D.Ion,
        ion: CoreProfiles.TimeSlice.Profiles1D.Ion,
        R0: float,
        rho_tor_boundary,
    ):
        """Convert flux to d,v ,
        @ref https://wpcd-workflows.github.io/ets.html#ds-and-vs-from-turbulence-codes-to-transport-solvers
        """
        inv_Ln = 1 / R0  # np.max(1 / R0, ion.density.dln / rho_tor_boundary)
        inv_LT = 1 / R0  # np.max(1 / R0, ion.temperature.dln / rho_tor_boundary)
        D_ = np.abs(spec.particles.flux) / inv_Ln
        Chi_ = np.abs(spec.energy.flux) / inv_LT
        D = np.maximum(D_, Chi_ / 5)
        Chi = np.maximum(Chi_, D_ / 5)
        spec.particles.d = D
        spec.particles.v = spec.particles.flux + D * ion.density.dln / rho_tor_boundary
        spec.energy.d = Chi
        spec.energy.v = spec.energy.flux + Chi * ion.temperature.dln / rho_tor_boundary


@sp_tree
class CoreTransport(IDS):
    Model = CoreTransportModel

    model: AoS[CoreTransportModel]

    def initialize(self, *args, **kwargs) -> None:
        super().initialize(*args, **kwargs)
        for model in self.model:
            model.initialize()

    def refresh(self, *args, equilibrium: Equilibrium = None, core_profiles: CoreProfiles = None, time=None, **kwargs):
        super().refresh(*args, time=time, **kwargs)

        for model in self.model:
            model.refresh(time=self.time, equilibrium=equilibrium, core_profiles=core_profiles, **kwargs)

    def flush(self):
        super().flush()
        for model in self.model:
            model.flush()



from __future__ import annotations
import numpy as np
from typing_extensions import Self

from spdm.core.aos import AoS
from spdm.core.expression import Expression
from spdm.core.sp_property import sp_property
from spdm.core.time_series import TimeSeriesAoS, TimeSlice
from spdm.core.geo_object import GeoObject, GeoObjectSet
from spdm.core.mesh import Mesh

from spdm.geometry.curve import Curve
from spdm.geometry.point import Point

from spdm.utils.tags import _not_found_

from .utilities import *

from ..utils.logger import logger
from ..ontology import equilibrium


@sp_tree(domain="grid")
class EquilibriumCoordinateSystem(equilibrium._T_equilibrium_coordinate_system):
    grid_type: Identifier

    grid: Mesh

    r: Field = sp_property(units="m")

    z: Field = sp_property(units="m")

    jacobian: Field = sp_property(units="mixed")

    tensor_covariant: array_type = sp_property(coordinate3="1...3", coordinate4="1...3", units="mixed")

    tensor_contravariant: array_type = sp_property(coordinate3="1...3", coordinate4="1...3", units="mixed")


@sp_tree
class EquilibriumGlobalQuantities(equilibrium._T_equilibrium_global_quantities):
    psi_axis: float = sp_property(units="Wb")

    psi_boundary: float = sp_property(units="Wb")

    b_field_tor_axis: float = sp_property(units="T")

    magnetic_axis: Point
    """ Magnetic axis position and toroidal field"""

    ip: float = sp_property(units="A")

    beta_pol: float

    beta_tor: float

    beta_normal: float

    li_3: float

    volume: float = sp_property(units="m^3")

    area: float = sp_property(units="m^2")

    surface: float = sp_property(units="m^2")

    length_pol: float = sp_property(units="m")

    @sp_tree
    class CurrentCentre:
        r: float = sp_property(units="m")
        z: float = sp_property(units="m")
        velocity_z: float = sp_property(units="m.s^-1")

    current_centre: CurrentCentre

    q_axis: float

    q_95: float

    @sp_tree
    class Qmin:
        value: float
        rho_tor_norm: float

    q_min: Qmin

    energy_mhd: float = sp_property(units="J")

    psi_external_average: float = sp_property(units="Wb")

    v_external: float = sp_property(units="V")

    plasma_inductance: float = sp_property(units="H")

    plasma_resistance: float = sp_property(units="ohm")


@sp_tree(domain="psi_norm")
class EquilibriumProfiles1D(equilibrium._T_equilibrium_profiles_1d):
    """
    1D profiles of the equilibrium quantities
    NOTE:
        - psi_norm is the normalized poloidal flux
        - psi is the poloidal flux,
        - 以psi_norm为主坐标, 是因为 psi_norm 时必定单调增的，而 psi 由于符号的原因，不一定时单调增的。
          scipy.interpolate 在一维插值时，要求 x 为单增。以 psi 为 磁面坐标，在插值时会造成问题。
          profiles_1d 中涉及对磁面坐标的求导和积分时需注意修正 ！！！！

    """

    @property
    def _root(self):  # -> EquilibriumTimeSlice:
        return self._parent

    @sp_property
    def grid(self) -> CoreRadialGrid:
        L = (self.psi[-1] - self.psi[0]) / (self.psi_norm[-1] - self.psi_norm[0])
        psi_axis = self.psi[-1] - L * self.psi_norm[-1]
        psi_boundary = L + psi_axis
        psi_norm = self.psi_norm
        rho_tor_norm = _not_found_
        rho_tor_boundary = _not_found_
        if self.rho_tor_norm is not _not_found_:
            rho_tor_norm = self.rho_tor_norm(self.psi_norm)
            if self.rho_tor is not _not_found_:
                rho_tor_boundary = self.rho_tor[-1] / self.rho_tor_norm[-1]

        return CoreRadialGrid(
            {
                "psi_norm": psi_norm,
                "rho_tor_norm": rho_tor_norm,
                "psi_axis": psi_axis,
                "psi_boundary": psi_boundary,
                "rho_tor_boundary": rho_tor_boundary,
            }
        )

    psi_norm: array_type | Expression = sp_property(units="-", label=r"\bar{\psi}")

    psi: array_type | Expression = sp_property(units="Wb", label=r"\psi")

    dphi_dpsi: Expression = sp_property(label=r"\frac{d\phi}{d\psi}", units="-")

    phi: Expression = sp_property(units="Wb", label=r"\phi")

    q: Expression = sp_property(units="-", label="q")

    pressure: Expression = sp_property(units="Pa", label="P")

    dpressure_dpsi: Expression = sp_property(units="Pa.Wb^-1", label=r"\frac{dP}{d\psi}")

    f: Expression = sp_property(units="T.m")

    f_df_dpsi: Expression = sp_property(units="T^2.m^2/Wb", label=r"\frac{f d f}{d \psi}")

    j_tor: Expression = sp_property(units="A \cdot m^{-2}")

    j_parallel: Expression = sp_property(units="A/m^2")

    magnetic_shear: Expression = sp_property(units="-")

    r_inboard: Expression = sp_property(units="m")

    r_outboard: Expression = sp_property(units="m")

    rho_tor: Expression = sp_property(units="m", label=r"\rho_{tor}")

    rho_tor_norm: Expression = sp_property(units="m", label=r"\bar{\rho_{tor}}")

    dpsi_drho_tor: Expression = sp_property(units="Wb/m", label=r"\frac{d\psi}{d\rho_{tor}}")

    @sp_property
    def geometric_axis(self) -> Point:
        return (self.major_radius, self.magnetic_z)

    minor_radius: Expression = sp_property(units="m")

    major_radius: Expression = sp_property(units="m")  # R0

    magnetic_z: Expression = sp_property(units="m")  # Z0

    elongation: Expression

    triangularity_upper: Expression

    triangularity_lower: Expression

    triangularity: Expression

    squareness_upper_inner: Expression

    squareness_upper_outer: Expression

    squareness_lower_inner: Expression

    squareness_lower_outer: Expression

    squareness: Expression = sp_property(default_value=zero)

    volume: Expression = sp_property(units="m^3")

    rho_volume_norm: Expression

    dvolume_dpsi: Expression = sp_property(units="m^3 \cdot Wb^{-1}")

    dvolume_drho_tor: Expression = sp_property(units="m^2", label=r"V^{\prime}")

    area: Expression = sp_property(units="m^2")

    darea_dpsi: Expression = sp_property(units="m^2 \cdot Wb^{-1}")

    darea_drho_tor: Expression = sp_property(units="m")

    surface: Expression = sp_property(units="m^2")

    trapped_fraction: Expression

    gm1: Expression
    gm2: Expression
    gm3: Expression
    gm4: Expression
    gm5: Expression
    gm6: Expression
    gm7: Expression
    gm8: Expression
    gm9: Expression

    b_field_average: Expression = sp_property(units="T")

    b_field_min: Expression = sp_property(units="T")

    b_field_max: Expression = sp_property(units="T")

    beta_pol: Expression

    mass_density: Expression = sp_property(units="kg \cdot m^{-3}")


@sp_tree(domain="grid")
class EquilibriumProfiles2D(equilibrium._T_equilibrium_profiles_2d):
    type: Identifier

    grid: Mesh

    r: Field = sp_property(units="m")

    z: Field = sp_property(units="m")

    psi: Field = sp_property(units="Wb")

    theta: Field = sp_property(units="rad")

    phi: Field = sp_property(units="Wb")

    j_tor: Field = sp_property(units="A.m^-2")

    j_parallel: Field = sp_property(units="A.m^-2")

    b_field_r: Field = sp_property(units="T")

    b_field_z: Field = sp_property(units="T")

    b_field_tor: Field = sp_property(units="T")


@sp_tree
class EquilibriumBoundary(equilibrium._T_equilibrium_boundary):
    type: int

    outline: Curve

    psi_norm: float = sp_property(default_value=0.995)

    psi: float = sp_property(units="Wb")

    geometric_axis: Point

    minor_radius: float = sp_property(units="m")

    elongation: float

    elongation_upper: float

    elongation_lower: float

    triangularity: float

    triangularity_upper: float

    triangularity_lower: float

    squareness_upper_inner: float

    squareness_upper_outer: float

    squareness_lower_inner: float

    squareness_lower_outer: float

    x_point: AoS[Point]

    strike_point: AoS[Point]

    active_limiter_point: Point


@sp_tree
class EquilibriumBoundarySeparatrix(equilibrium._T_equilibrium_boundary_separatrix):
    type: int

    outline: GeoObjectSet

    psi: float = sp_property(units="Wb")

    geometric_axis: Point

    minor_radius: float = sp_property(units="m")

    elongation: float

    elongation_upper: float

    elongation_lower: float

    triangularity: float

    triangularity_upper: float

    triangularity_lower: float

    squareness_upper_inner: float

    squareness_upper_outer: float

    squareness_lower_inner: float

    squareness_lower_outer: float

    x_point: AoS[Point]

    strike_point: AoS[Point]

    active_limiter_point: Point


@sp_tree
class EequilibriumConstraints(equilibrium._T_equilibrium_constraints):
    pass


@sp_tree
class EquilibriumGGD(equilibrium._T_equilibrium_ggd):
    pass


@sp_tree
class EquilibriumTimeSlice(equilibrium._T_equilibrium_time_slice):
    vacuum_toroidal_field: VacuumToroidalField

    Boundary = EquilibriumBoundary
    boundary: EquilibriumBoundary

    BoundarySeparatrix = EquilibriumBoundarySeparatrix
    boundary_separatrix: EquilibriumBoundarySeparatrix

    Constraints = EequilibriumConstraints
    constraints: EequilibriumConstraints

    GlobalQuantities = EquilibriumGlobalQuantities
    global_quantities: EquilibriumGlobalQuantities

    Profiles1D = EquilibriumProfiles1D
    profiles_1d: EquilibriumProfiles1D

    Profiles2D = EquilibriumProfiles2D
    profiles_2d: EquilibriumProfiles2D

    CoordinateSystem = EquilibriumCoordinateSystem
    coordinate_system: EquilibriumCoordinateSystem

    GGD = EquilibriumGGD
    ggd: GGD

    def __view__(self, view_point="RZ", **kwargs):
        """
        plot o-point,x-point,lcfs,separatrix and contour of psi
        """

        geo = {}

        match view_point.lower():
            case "rz":
                geo["psi"] = self.profiles_2d.psi.__view__()

                try:
                    geo["o_points"] = Point(
                        self.global_quantities.magnetic_axis.r,
                        self.global_quantities.magnetic_axis.z,
                        styles={"$matplotlib": {"color": "red", "marker": ".", "linewidths": 0.5}},
                    )

                    geo["x_points"] = [
                        Point(
                            p.r,
                            p.z,
                            name=f"{idx}",
                            styles={"$matplotlib": {"color": "blue", "marker": "x", "linewidths": 0.5}},
                        )
                        for idx, p in enumerate(self.boundary.x_point)
                    ]

                    # geo["strike_points"] = [
                    #     Point(p.r, p.z, name=f"{idx}") for idx, p in enumerate(self.boundary.strike_point)
                    # ]

                    geo["boundary"] = self.boundary.outline
                    geo["boundary"]._metadata["styles"] = {
                        "$matplotlib": {"color": "blue", "linestyle": "dotted", "linewidth": 0.5}
                    }
                    # geo["boundary_separatrix"] = self.boundary_separatrix.outline
                    # if geo["boundary_separatrix"] is not _not_found_:
                    #     geo["boundary_separatrix"]._metadata["styles"] = {
                    #         "$matplotlib": {"color": "red", "linestyle": "dashed", "linewidth": 0.25}
                    #     }
                except Exception as error:
                    raise error

        geo["styles"] = kwargs

        return geo

    # def __view__(self, view_port="RZ", **kwargs) -> GeoObject:
    #     geo = {}

    #     if view_port == "RZ":
    #         o_points, x_points = self.coordinate_system.critical_points

    #         geo["o_points"] = [Point(p.r, p.z, name=f"{idx}") for idx, p in enumerate(o_points)]
    #         geo["x_points"] = [Point(p.r, p.z, name=f"{idx}") for idx, p in enumerate(x_points)]

    #         geo["boundary"] = Curve(self.boundary.outline.r.__array__(), self.boundary.outline.z.__array__())

    #         geo["boundary_separatrix"] = Curve(
    #             self.boundary_separatrix.outline.r.__array__(),
    #             self.boundary_separatrix.outline.z.__array__(),
    #         )

    #     geo["psi"] = self.profiles_2d.psi.__view__()

    #     styles = {
    #         "o_points": {"$matplotlib": {"c": "red", "marker": "."}},
    #         "x_points": {"$matplotlib": {"c": "blue", "marker": "x"}},
    #         "boundary": {"$matplotlib": {"color": "red", "linewidth": 0.5}},
    #         "boundary_separatrix": {
    #             "$matplotlib": {
    #                 "color": "red",
    #                 "linestyle": "dashed",
    #                 "linewidth": 0.25,
    #             }
    #         },
    #     }
    #     styles = update_tree(styles, kwargs)

    #     return geo, styles


from .wall import Wall
from .tf import TF
from .magnetics import Magnetics
from .pf_active import PFActive


@sp_tree
class Equilibrium(IDS):
    r"""
    Description of a 2D, axi-symmetric, tokamak equilibrium; result of an equilibrium code.

    Reference:

    - O. Sauter and S. Yu Medvedev, "Tokamak coordinate conventions: COCOS", Computer Physics Communications 184, 2 (2013), pp. 293--302.
    """

    code: Code = {"name": "fy_eq"}  # default plugin

    ids_properties: IDSProperties

    TimeSlice = EquilibriumTimeSlice

    time_slice: TimeSeriesAoS[EquilibriumTimeSlice]

    def __view__(self, *args, **kwargs):
        current = self.time_slice.current
        return current.__view__(*args, **kwargs) if current is not _not_found_ else {}

    def refresh(
        self,
        *args,
        time=None,
        wall: Wall = None,
        tf: TF = None,
        magnetics: Magnetics = None,
        pf_active: PFActive = None,
        **kwargs,
    ) -> None:
        return super().refresh(
            *args,
            time=time,
            tf=tf,
            wall=wall,
            magnetics=magnetics,
            pf_active=pf_active,
            **kwargs,
        )


r"""
  COCOS  11
    ```{text}
        Top view
                ***************
                *               *
            *   ***********   *
            *   *           *   *
            *   *             *   *
            *   *             *   *
        Ip  v   *             *   ^  \phi
            *   *    Z o--->R *   *
            *   *             *   *
            *   *             *   *
            *   *     Bpol    *   *
            *   *     o     *   *
            *   ***********   *
                *               *
                ***************
                Bpol x
                Poloidal view
            ^Z
            |
            |       ************
            |      *            *
            |     *         ^    *
            |     *   \rho /     *
            |     *       /      *
            +-----*------X-------*---->R
            |     *  Ip, \phi   *
            |     *              *
            |      *            *
            |       *****<******
            |       Bpol,\theta
            |
                Cylindrical coordinate      : $(R,\phi,Z)$
        Poloidal plane coordinate   : $(\rho,\theta,\phi)$
    ```
"""



import numpy as np
import scipy.constants
from spdm.core.sp_property import sp_tree
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_transport import CoreTransport
from fytok.modules.core_sources import CoreSources
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.utilities import *

PI = scipy.constants.pi


@sp_tree
class PredefinedSource(CoreSources.Source):
    identifier = "predefined"
    code = {"name": "predefined", "description": f"predefined"}

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D, *args, **kwargs) -> CoreSources.Source.TimeSlice:
        current: CoreSources.Source.TimeSlice = super().fetch(*args, **kwargs)

        rho_tor_norm = profiles_1d.rho_tor_norm

        _x = rho_tor_norm

        S = 9e20 * np.exp(15.0 * (_x**2 - 1.0))

        current.profiles_1d.grid = profiles_1d.grid
        current.profiles_1d.electrons.particles = S

        current.profiles_1d.ion.extend(
            [
                {"@name": "D", "particles": S * 0.5},
                {"@name": "T", "particles": S * 0.5},
                # {"@name": "He", "particles": S * 0.02},
            ]
        )

        return current


CoreSources.Source.register(["predefined"], PredefinedSource)



__path__ = __import__("pkgutil").extend_path(__path__, __name__)

import os

from .ontology import GLOBAL_ONTOLOGY

from .utils.envs import *

__version__ = FY_VERSION

############################################################


try:
    from importlib import resources as impresources
    from . import _mapping
    from spdm.core.entry import EntryProxy

    EntryProxy._mapping_path.extend([p.resolve() for p in impresources.files(_mapping)._paths])

except Exception as error:
    raise FileNotFoundError(f"Can not find mappings!") from error

############################################################


if FY_VERBOSE != "quiet":
    from spdm.utils.logger import logger
    from spdm.utils.envs import SP_MPI

    if SP_MPI is None or SP_MPI.COMM_WORLD.Get_rank() == 0:  # 粗略猜测是否在交互环境下运行
        logger.info(FY_LOGO)



import os
import pathlib

import numpy as np
from spdm.core.entry import Entry
from spdm.core.file import File
from spdm.view.sp_view import display

from fytok.modules.equilibrium import Equilibrium
from fytok.tokamak import Tokamak

from fytok.utils.logger import logger

WORKSPACE = "/home/salmon/workspace"  # "/ssd01/salmon_work/workspace/"

os.environ["SP_DATA_MAPPING_PATH"] = f"{WORKSPACE}/fytok_data/mapping"


if __name__ == "__main__":
    output_path = pathlib.Path(f"{WORKSPACE}/output/")

    shot = 70745

    tok = Tokamak(C)

    tok.refresh(time=9.26)

    psi = tok.equilibrium.time_slice.current.profiles_2d[0].psi.__array__()

    psi_min = psi.min()

    psi_max = psi.max()

    levels = np.arange(psi_min, psi_max, (psi_max-psi_min)/40)

    for i in range(150):
        display(tok.equilibrium,
                title=f"EAST shot={shot} time={tok.equilibrium.time_slice.current.time:.2f}s ",
                output=output_path / f"{tok.tag}.png",
                transparent=False,
                styles={"psi": {"$matplotlib": {"levels": levels}}}
                )
        logger.debug(f"Equilibrium [{i:5d}] time={tok.equilibrium.time_slice.current.time:.2f}s")

        tok.advance()

    # convert -delay 10 -loop 1 tok_east*.png tok_east_70754.mp4

    logger.debug(tok.equilibrium.time_slice.time)



from functools import cached_property

from spdm.utils.tags import _undefined_
from  ..ontology import edge_transport import _T_edge_transport, _T_edge_transport_model


class EdgeTransport(_T_edge_transport):

    Model = _T_edge_transport_model

    @cached_property
    def model_combiner(self) -> Model:
        return self.model.combine({
            "identifier": {"name": "combined", "index": 1,
                           "description": """Combination of data from all available transport models"""},
            "code": {"name": _undefined_}
        })

    def update(self, *args, **kwargs) -> float:
        if "model_combiner" in self.__dict__:
            del self.__dict__["model_combiner"]
        return sum([model.refresh(*args,   **kwargs) for model in self.model])




from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_transport import CoreTransport
from fytok.modules.equilibrium import Equilibrium


class ChangHiton(CoreTransport.Model):
    """
        Chang-Hiton formula for \Chi_{i}
        ===============================

        References:
        =============
        - Tokamaks, Third Edition, Chapter 14.11  ,p737,  J.A.Wesson 2003
    """

    def refresh(self, *args,
                core_profiles_1d: CoreProfiles.Profiles1d = None,
                equilibrium: Equilibrium.TimeSlice = None,
                **kwargs):
        return super().refresh(*args, **kwargs)


__SP_EXPORT__ = ChangHiton



import collections
import collections.abc

import numpy as np
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_transport import CoreTransport
from fytok.modules.equilibrium import Equilibrium
from scipy import constants
from spdm.core.function import function_like
from spdm.numlib.misc import array_like
from fytok.utils.logger import logger


class Spitzer(CoreTransport.Model):
    """
    Spitzer Resistivity

    References:
    - Tokamaks, Third Edition, Chapter 14  ,p727,  J.A.Wesson 2003
    """

    _metadata = {
        "identifier": "neoclassical",
        "code": {
            "name": "spitzer",
            "description": f" Spitzer Resistivity",
        },
    }

    def refresh(
        self, *args, equilibrium: Equilibrium.TimeSlice, core_profiles_1d: CoreProfiles.Profiles1d, **kwargs
    ) -> float:
        # residual = super().refresh(*args, equilibrium=equilibrium, core_profiles=core_profiles, **kwargs)

        eV = constants.electron_volt

        radial_grid = core_profiles_1d.grid

        B0 = radial_grid.b0
        R0 = radial_grid.r0

        rho_tor_norm = radial_grid.rho_tor_norm
        rho_tor = radial_grid.rho_tor
        psi_norm = radial_grid.psi_norm(rho_tor)
        psi_axis = equilibrium.global_quantities.psi_axis
        psi_boundary = equilibrium.global_quantities.psi_boundary
        psi = psi_norm * (psi_boundary - psi_axis) + psi_axis

        q = equilibrium.profiles_1d.q(psi)

        # Tavg = np.sum([ion.density*ion.temperature for ion in core_profile.ion]) / \
        #     np.sum([ion.density for ion in core_profile.ion])

        Te = core_profiles_1d.electrons.temperature(rho_tor_norm)
        Ne = core_profiles_1d.electrons.density(rho_tor_norm)
        # Pe = core_profile.electrons.pressure(rho_tor_norm)

        # Coulomb logarithm
        #  Ch.14.5 p727 Tokamaks 2003
        # lnCoul = (14.9 - 0.5*np.log(Ne/1e20) + np.log(Te/1000)) * (Te < 10) +\
        #     (15.2 - 0.5*np.log(Ne/1e20) + np.log(Te/1000))*(Te >= 10)
        # (17.3 - 0.5*np.log(Ne/1e20) + 1.5*np.log(Te/1000))*(Te >= 10)

        # lnCoul = 14
        lnCoul = core_profiles_1d.coulomb_logarithm(rho_tor_norm)

        # electron collision time , eq 14.6.1
        tau_e = 1.09e16 * ((Te / 1000) ** (3 / 2)) / Ne / lnCoul

        vTe = np.sqrt(Te * eV / constants.electron_mass)

        # Larmor radius,   eq 14.7.2
        # rho_e = 1.07e-4*((Te/1000)**(1/2))/B0

        # rho_tor[0] = max(rho_e[0], rho_tor[0])

        epsilon = rho_tor / R0
        epsilon12 = np.sqrt(epsilon)
        epsilon32 = epsilon ** (3 / 2)
        ###########################################################################################
        #  Sec 14.10 Resistivity
        #
        eta_s = 1.65e-9 * lnCoul * (Te / 1000) ** (-3 / 2)
        Zeff = core_profiles_1d.zeff(rho_tor_norm)
        fT = 1.0 - (1 - epsilon) ** 2 / np.sqrt(1.0 - epsilon**2) / (1 + 1.46 * np.sqrt(epsilon))

        phi = np.zeros_like(rho_tor_norm)
        nu_e = R0 * q[1:] / vTe[1:] / tau_e[1:] / epsilon32[1:]
        phi[1:] = fT[1:] / (1.0 + (0.58 + 0.20 * Zeff[1:]) * nu_e)
        phi[0] = 0

        C = 0.56 / Zeff * (3.0 - Zeff) / (3.0 + Zeff)

        eta = eta_s * Zeff / (1 - phi) / (1.0 - C * phi) * (1.0 + 0.27 * (Zeff - 1.0)) / (1.0 + 0.47 * (Zeff - 1.0))

        self.profiles_1d.current["conductivity_parallel"] = function_like(
            array_like(rho_tor_norm, 1.0 / eta), rho_tor_norm
        )


__SP_EXPORT__ = Spitzer



from __future__ import annotations

from spdm.core.geo_object import GeoObject

from ..ontology import ec_launchers


class ECLaunchers(ec_launchers._T_ec_launchers):
    def __view__(self, view_point="RZ", **kwargs) -> GeoObject:
        geo = {}
        styles = {}
        match view_point.lower():
            case "top":
                geo["beam"] = [beam.name for beam in self.beam]
                styles["beam"] = {"$matplotlib": {"color": "blue"}, "text": True}

        return geo



from __future__ import annotations

from scipy import constants
from copy import copy
import math
from spdm.core.expression import Expression, Variable, zero
from spdm.core.sp_property import sp_tree, sp_property, PropertyTree
from spdm.core.time_series import TimeSlice, TimeSeriesAoS
from spdm.core.aos import AoS
from spdm.utils.tags import _not_found_
from spdm.utils.typing import array_type

from .core_profiles import CoreProfiles
from .core_sources import CoreSources
from .core_transport import CoreTransport
from .equilibrium import Equilibrium
from .utilities import *
from ..utils.logger import logger

# from ..ontology import transport_solver_numerics

EPSILON = 1.0e-15
TOLERANCE = 1.0e-6

TWOPI = 2.0 * constants.pi


@sp_tree
class TransportSolverNumericsEquation:
    """Profile and derivatives a the primary quantity for a 1D transport equation"""

    identifier: str = sp_property(alias="@name")
    """ Identifier of the primary quantity of the transport equation. The description
        node contains the path to the quantity in the physics IDS (example:
        core_profiles/profiles_1d/ion/D/density)"""

    profile: array_type | Expression
    """ Profile of the primary quantity"""

    flux: array_type | Expression
    """ Flux of the primary quantity"""

    units: typing.Tuple[float, float]

    d_dr: array_type | Expression
    """ Radial derivative with respect to the primary coordinate"""

    dflux_dr: array_type | Expression
    """ Radial derivative of Flux of the primary quantity"""

    d2_dr2: array_type | Expression
    """ Second order radial derivative with respect to the primary coordinate"""

    d_dt: array_type | Expression
    """ Time derivative"""

    d_dt_cphi: array_type | Expression
    """ Derivative with respect to time, at constant toroidal flux (for current
        diffusion equation)"""

    d_dt_cr: array_type | Expression
    """ Derivative with respect to time, at constant primary coordinate coordinate (for
        current diffusion equation)"""

    coefficient: typing.List[typing.Any]
    """ Set of numerical coefficients involved in the transport equation
       
        [d_dt,D,V,RHS]
        
        d_dt + flux'= RHS  
        
        flux =-D y' + V y

        u * y + v* flux - w =0 
    """

    boundary_condition_type: int = 1

    boundary_condition_value: tuple
    """ [u,v,v] 
    
    u * profile + v* flux - w =0"""

    convergence: PropertyTree
    """ Convergence details"""


@sp_tree(coordinate1="grid/rho_tor_norm")
class TransportSolverNumericsTimeSlice(TimeSlice):
    """Numerics related to 1D radial solver for a given time slice"""

    grid: CoreRadialGrid

    equations: AoS[TransportSolverNumericsEquation]
    """ Set of transport equations"""

    control_parameters: PropertyTree
    """ Solver-specific input or output quantities"""

    drho_tor_dt: array_type | Expression = sp_property(units="m.s^-1")
    """ Partial derivative of the toroidal flux coordinate profile with respect to time"""

    d_dvolume_drho_tor_dt: array_type | Expression = sp_property(units="m^2.s^-1")
    """ Partial derivative with respect to time of the derivative of the volume with
      respect to the toroidal flux coordinate"""


@sp_tree
class TransportSolverNumerics(IDS):
    r"""Solve transport equations  $\rho=\sqrt{ \Phi/\pi B_{0}}$"""

    code: Code = {"name": "fy_trans"}

    solver: str = "ion_solver"

    ion_thermal: set = set()

    ion_non_thermal: set = set()

    impurities: set = set()

    neutral: set = set()

    primary_coordinate: str = "rho_tor_norm"
    r""" 与 core_profiles 的 primary coordinate 磁面坐标一致
      rho_tor_norm $\bar{\rho}_{tor}=\sqrt{ \Phi/\Phi_{boundary}}$ """

    equations: AoS[TransportSolverNumericsEquation] = []

    variables: typing.Dict[str, Expression] = {}

    profiles_1d: CoreProfiles.TimeSlice.Profiles1D = {}

    TimeSlice = TransportSolverNumericsTimeSlice

    time_slice: TimeSeriesAoS[TransportSolverNumericsTimeSlice] = []

    def initialize(self, *args, **kwargs):
        return super().initialize(*args, **kwargs)

    def preprocess(self, *args, **kwargs) -> TransportSolverNumericsTimeSlice:
        eq_grid: CoreRadialGrid = self.inports["equilibrium/time_slice/0/profiles_1d/grid"].fetch()

        rho_tor_norm = kwargs.get("rho_tor_norm", _not_found_)

        if rho_tor_norm is _not_found_:
            rho_tor_norm = self.code.parameters.rho_tor_norm

        new_grid = eq_grid.remesh(rho_tor_norm)

        self.profiles_1d["grid"] = new_grid

        current = super().preprocess(*args, **kwargs)

        current["grid"] = new_grid

        return current

    def execute(
        self,
        current: TransportSolverNumericsTimeSlice,
        *previous: TransportSolverNumericsTimeSlice,
    ) -> TransportSolverNumericsTimeSlice:
        logger.info(f"Solve transport equations : { '  ,'.join([equ.identifier for equ in self.equations])}")
        return super().execute(current, *previous)

    def postprocess(self, current: TransportSolverNumericsTimeSlice) -> TransportSolverNumericsTimeSlice:
        return super().postprocess(current)

    def refresh(
        self,
        *args,
        equilibrium: Equilibrium = None,
        core_transport: CoreTransport = None,
        core_sources: CoreSources = None,
        core_profiles: CoreProfiles = None,
        **kwargs,
    ) -> TransportSolverNumericsTimeSlice:
        return super().refresh(
            *args,
            equilibrium=equilibrium,
            core_transport=core_transport,
            core_sources=core_sources,
            core_profiles=core_profiles,
            **kwargs,
        )

    def fetch(self, *args, **kwargs) -> CoreProfiles.TimeSlice.Profiles1D:
        """获得 CoreProfiles.TimeSlice.Profiles1D 形式状态树。"""

        current: TransportSolverNumericsTimeSlice = self.time_slice.current

        X = current.grid.rho_tor_norm

        Y = sum([[equ.profile, equ.flux] for equ in current.equations], [])

        self.profiles_1d["grid"] = current.grid

        profiles_1d = self.profiles_1d.fetch(X, *Y)

        profiles_1d[self.primary_coordinate] = current.grid.get(self.primary_coordinate)

        return profiles_1d



import typing
import scipy.constants
from spdm.core.expression import Variable, zero
from spdm.core.sp_property import sp_tree
from fytok.utils.atoms import nuclear_reaction, atoms
from fytok.utils.logger import logger
from fytok.modules.core_sources import CoreSources
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.utilities import *

PI = scipy.constants.pi


@sp_tree
class FusionReaction(CoreSources.Source):
    """[summary]

    Args:
        CoreSources ([type]): [description]


    $\\alpha$输运模型参考[@angioniGyrokineticCalculationsDiffusive2008; @angioniGyrokineticSimulationsImpurity2009]

    * energetic $\\alpha$ particle density $n_{\\alpha}$

    $$
    \\frac{\\partial n_{\\alpha}}{\\partial t}+\\nabla\\left(-D_{\\alpha}\\nabla n_{\\alpha}+Vn_{\\alpha}\\right)=-\\frac{n_{\\alpha}}{\\tau_{sd}^{*}}+n_{D}n_{T}\\left\\langle \\sigma v\\right\\rangle _{DT}
    $$

    * $He$ ash density $n_{He}$

    $$
    \\frac{\\partial n_{He}}{\\partial t}+\\nabla\\left(-D_{He}\\nabla n_{He}+Vn_{He}\\right)=\\frac{n_{\\alpha}}{\\tau_{sd}^{*}}
    $$

    where
    $$
    \\tau_{sd}^{*}=\\ln\\left(v_{\\alpha}^{3}/v_{c}^{3}+1\\right)\\left(m_{e}m_{\\alpha}v_{e}^{3}\\right)/\\left(64\\sqrt{\\pi}e^{4}n_{e}\\ln\\Lambda\\right)
    $$
    is the actual thermalization slowing down time.

    energetic $\\alpha$ particle flux
    $$
    \\frac{R\\Gamma_{\\alpha}}{n_{\\alpha}}=D_{\\alpha}\\left(\\frac{R}{L_{n_{\\alpha}}}C_{p_{\\alpha}}\\right)
    $$
    where
    $$
    D_{\\alpha}=D_{\\text{He}}\\left[0.02+4.5\\left(\\frac{T_{e}}{E_{\\alpha}}\\right)+8\\left(\\frac{T_{e}}{E_{\\alpha}}\\right)^{2}+350\\left(\\frac{T_{e}}{E_{\\alpha}}\\right)^{3}\\right]
    $$
    and
    $$
    C_{p_{\\alpha}}=\\frac{3}{2}\\frac{R}{L_{T_{e}}}\\left[\\frac{1}{\\log\\left[\\left(E_{\\alpha}/E_{c}\\right)^{3/2}+1\\right]\\left[1+\\left(E_{c}/E_{\\alpha}\\right)^{3/2}\\right]}-1\\right]
    $$
    Here $E_{c}$ is the slowing down critical energy. We remind that $E_{c}/E_{\\alpha}=33.05 T_e/E_{\\alpha}$, where $E_{\\alpha}=3500 keV$  is the thirth energy of $\\alpha$ particles.
    """

    identifier = "fusion"

    code = {"name": "fusion", "description": "Fusion reaction"}  # type: ignore

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        x = np.linspace(0, 10.0, 256)
        _x = Variable(0, "x")
        self._sivukhin = Function(x, 1.0 / (1 + x**1.5)).I(_x) / (_x)
        self._sivukhin._metadata["name"] = "sivukhin"
        self._sivukhin._metadata["label"] = "F"

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D) -> CoreSources.Source.TimeSlice:
        current: CoreSources.Source.TimeSlice = super().fetch(profiles_1d)

        heating = self.code.parameters.heating is not False

        rho_tor_norm = profiles_1d.rho_tor_norm

        source_1d = current.profiles_1d
        m_e = scipy.constants.physical_constants["electron mass"][0]
        m_alpha = scipy.constants.physical_constants["alpha particle mass"][0]
        a_alpha = 4
        Te = profiles_1d.electrons.temperature

        fusion_reactions: typing.List[str] = self.code.parameters.fusion_reactions or []

        ne = profiles_1d.electrons.density

        for tag in fusion_reactions:
            if tag != "D(t,n)alpha":
                raise NotImplementedError(f"NOT IMPLEMENTED YET！！ By now only support D(t,n)alpha!")
            reaction = nuclear_reaction[tag]

            r0, r1 = reaction.reactants
            p0, p1 = reaction.products

            pa = atoms[p1].label

            n0 = profiles_1d.ion[r0].density
            n1 = profiles_1d.ion[r1].density

            T0 = profiles_1d.ion[r0].temperature
            T1 = profiles_1d.ion[r1].temperature
            ni = n0 + n1
            Ti = (n0 * T0 + n1 * T1) / ni
            nEP = profiles_1d.ion[p1].density

            E0, E1 = reaction.energy

            lnGamma = 17  # FIXME: 粗略估计

            C = zero
            a_tot = 0

            for label in [r0, r1]:
                ion = profiles_1d.ion[label]
                a_tot += ion.a
                C += ion.density * (ion.z**2) / (ion.a / a_alpha)

            C /= ne

            Ecrit = (Te) * (4 * np.sqrt(m_e / m_alpha) / (3 * np.sqrt(PI) * C)) ** (-2.0 / 3.0)

            # nu_slowing_down = (ni * 1.0e-19 * lnGamma) / (1.99 * ((Ti / 1000) ** (3 / 2)))

            tau_e = (1.99 * ((Te / 1000) ** (3 / 2))) / (ne * 1.0e-19 * lnGamma)

            tau_s = tau_e * np.log((E1 / Ecrit) ** 1.5 + 1) / 3

            S = reaction.reactivities(Ti) * n0 * n1

            if r0 == r1:
                S *= 0.5

            # r_ped = 0.99  #

            # delta = np.heaviside(rho_tor_norm - r_ped, 0.5)

            # S = S * (1 - delta)

            source_1d.ion[r0].particles -= S
            source_1d.ion[r1].particles -= S
            source_1d.ion[p0].particles += S
            source_1d.ion[p1].particles += S - nEP / tau_s
            source_1d.ion[pa].particles += nEP / tau_s

            Efus = S * E1  # nEP * nu_slowing_down

            frac = 0.0

            # if heating:
            # 离子加热分量
            #  [Stix, Plasma Phys. 14 (1972) 367 Eq.15

            frac = self._sivukhin(E1 / Ecrit) 

            # 加热离子
            for label in [r0, r1]:
                ion = source_1d.ion[label]

                ion.energy += Efus * frac * ion.a / a_tot

            source_1d.electrons.energy += Efus * (1.0 - frac)

        return current


CoreSources.Source.register(["fusion"], FusionReaction)



import collections
import collections.abc
import functools
import typing
import numpy as np
import scipy.constants
from dataclasses import dataclass


from spdm.core.htree import List
from spdm.core.sp_property import sp_property, sp_tree
from spdm.core.time_series import TimeSeriesAoS
from spdm.core.expression import Expression, Variable
from spdm.core.field import Field
from spdm.core.geo_object import GeoObject, GeoObjectSet
from spdm.geometry.point import Point
from spdm.geometry.curve import Curve
from spdm.core.mesh import Mesh
from spdm.mesh.mesh_curvilinear import CurvilinearMesh
from spdm.utils.tags import _not_found_
from spdm.utils.typing import ArrayLike, NumericType, array_type, scalar_type

from fytok.modules.equilibrium import Equilibrium
from fytok.modules.utilities import *
from fytok.utils.logger import logger

from .contours import find_critical_points, find_contours


PI = scipy.constants.pi

_R = Variable(0, "R")
_Z = Variable(1, "Z")


@dataclass
class OXPoint:
    r: float
    z: float
    psi: float


TOLERANCE = 1.0e-6


# fmt:off
COCOS_TABLE = [
    # e_Bp ,    $\sigma_{Bp}$,    $\sigma_{R\varphi\Z}$,  $\sigma_{\rho\theta\varphi}$
    None,                                                                                    # 0
    (0,         +1,                  +1,                       +1                   ),       # 1
    (0,         +1,                  -1,                       +1                   ),       # 2
    (0,         -1,                  +1,                       -1                   ),       # 3
    (0,         -1,                  -1,                       -1                   ),       # 4
    (0,         +1,                  +1,                       -1                   ),       # 5
    (0,         +1,                  -1,                       -1                   ),       # 6
    (0,         -1,                  +1,                       +1                   ),       # 7
    (0,         -1,                  -1,                       +1                   ),       # 8
    None,                                                                                    # 9
    None,                                                                                    # 10
    (1,         +1,                  +1,                       +1                   ),       # 11
    (1,         +1,                  -1,                       +1                   ),       # 12
    (1,         -1,                  +1,                       -1                   ),       # 13
    (1,         -1,                  -1,                       -1                   ),       # 14
    (1,         +1,                  +1,                       -1                   ),       # 15
    (1,         +1,                  -1,                       -1                   ),       # 16
    (1,         -1,                  +1,                       +1                   ),       # 17
    (1,         -1,                  -1,                       +1                   ),       # 18
]
# fmt:on


@sp_tree
class FyEquilibriumCoordinateSystem(Equilibrium.TimeSlice.CoordinateSystem):
    r"""
    Flux surface coordinate system on a square grid of flux and poloidal angle
    默认采用磁面坐标

    $$
        V^{\prime}\left(\rho\right)=\frac{\partial V}{\partial\rho}=2\pi\int_{0}^{2\pi}\sqrt{g}d\theta=2\pi\oint\frac{R}{\left|\nabla\rho\right|}dl
    $$

    $$
        \left\langle\alpha\right\rangle\equiv\frac{2\pi}{V^{\prime}}\int_{0}^{2\pi}\alpha\sqrt{g}d\theta=\frac{2\pi}{V^{\prime}}\varoint\alpha\frac{R}{\left|\nabla\rho\right|}dl
    $$

    Magnetic Flux Coordinates
    psi         :                     ,  flux function , $B \cdot \nabla \psi=0$ need not to be the poloidal flux funcion $\Psi$
    theta       : 0 <= theta   < 2*pi ,  poloidal angle
    phi         : 0 <= phi     < 2*pi ,  toroidal angle
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._s_B0 = np.sign(self.b0)
        self._s_Ip = np.sign(self.ip)

        self._e_Bp, self._s_Bp, self._s_RpZ, self._s_rtp = COCOS_TABLE[5]

        self._s_eBp_2PI = 1.0 if self._e_Bp == 0 else (2.0 * scipy.constants.pi)

        self.psirz = self._parent.profiles_2d.psi

        o_points, x_points = find_critical_points(self.psirz)

        if len(o_points) == 0:
            raise RuntimeError(f"Can not find O-point!")

        if len(x_points) == 0:
            raise RuntimeError(f"Can not find X-point!")

        self.x_point = x_points

        self.magnetic_axis = Point(o_points[0].r, o_points[0].z)

        self.psi_axis = o_points[0].value

        self.psi_boundary = x_points[0].value

        # 磁面坐标
        self.psi_norm = self._parent.profiles_1d.psi_norm

        if self.psi_norm is _not_found_ or self.psi_norm is None:
            self.psi_norm = np.linspace(0.001, 0.9995, 64)
            # logger.warning(f"Can not get psi_norm, using default {psi_norm}")

        elif not np.all(np.diff(self.psi_norm) > 0):
            raise RuntimeError(f"psi_norm is not monotonically increasing!")

        # if np.isclose(self.psi_norm[0], 0.0) and np.isclose(self.psi_norm[-1], 1.0):
        #     logger.warning(
        #         f"Singular values are caused when psi_norm takes values of 0.0 or 1.0.! {self.psi_norm[0]} {self.psi_norm[-1]}"
        #     )

        # 磁面坐标的函数，ffprime，pprime

    b0: float = sp_property(alias="../vacuum_toroidal_field/b0")
    r0: float = sp_property(alias="../vacuum_toroidal_field/r0")
    ip: float = sp_property(alias="../global_quantities/ip")

    # magnetic_axis: Point = sp_property(alias="../global_quantities/magnetic_axis")
    # psi_axis: float = sp_property(alias="../global_quantities/psi_axis")
    # psi_boundary: float = sp_property(alias="../global_quantities/psi_boundary")
    # psi_norm: array_type = sp_property(alias="../profiles_1d/psi_norm")

    ffprime: Expression = sp_property(alias="../profiles_1d/f_df_dpsi", label=r" \frac{f df}{d\psi}")
    pprime: Expression = sp_property(alias="../profiles_1d/dpressure_dpsi", label=r" \frac{dP}{d\psi}")

    # psirz: Field = sp_property(alias="../profiles_2d/psi")

    @sp_property
    def dvolume_dpsi(self) -> Expression:
        return Function(*self._surface_integral(1.0), name="dvolume_dpsi", label=r"\frac{d V}{d\psi}")

    @sp_property
    def Bpol(self) -> Expression:
        r"""$B_{pol}= \left|\nabla \psi \right|/2 \pi R $"""
        return (
            np.sqrt(self.psirz.pd(1, 0) ** 2 + self.psirz.pd(0, 1) ** 2)
            / _R
            * np.abs(self._s_RpZ * self._s_Bp / self._s_eBp_2PI)
        )

    #################################
    # Profiles 2D

    @sp_property
    def grid(self) -> Mesh:
        theta = self.get_cache("grid/dim2", _not_found_)

        if theta is _not_found_:
            ntheta = self.get_cache(".../code/parameters/num_of_theta", 64)
            theta = np.linspace(0, 2.0 * scipy.constants.pi, ntheta, endpoint=False)

        if not (isinstance(theta, np.ndarray) and theta.ndim == 1):
            raise ValueError(f"Can not create grid! theta={theta}")

        psi_norm = self.get_cache("grid/dim1", _not_found_)

        if psi_norm is _not_found_:
            psi_norm = self.psi_norm

        if psi_norm is _not_found_:
            psi_norm = self.get_cache(".../code/parameters/psi_norm", np.linspace(0.0, 0.995, 128))

        surfs = GeoObjectSet([surf for _, surf in self.find_surfaces(psi_norm)])

        return CurvilinearMesh(psi_norm, theta, geometry=surfs, mesh={"periods": [False, 2.0 * scipy.constants.pi]})

    @sp_property(mesh="grid")
    def r(self) -> Field:
        return self.grid.points[0]

    @sp_property(mesh="grid")
    def z(self) -> Field:
        return self.grid.points[1]

    @sp_property(mesh="grid")
    def jacobian(self) -> Field:
        raise NotImplementedError(f"")

    @sp_property(mesh="grid")
    def tensor_covariant(self) -> Field:
        raise NotImplementedError(f"")

    @sp_property(mesh="grid")
    def tensor_contravariant(self) -> Field:
        raise NotImplementedError(f"")

    ###############################
    # surface integral
    def find_surfaces_by_psi(self, psi) -> typing.Generator[typing.Tuple[float, GeoObject], None, None]:

        psi_axis = self.psi_axis
        psi_boundary = self.psi_boundary
        R = self.magnetic_axis.r
        Z = self.magnetic_axis.z

        for psi_val, surfs in find_contours(self.psirz, values=psi):

            if np.isclose(psi_val, psi_axis):
                yield psi_val, Point(R, Z)

            # elif np.isclose(psi_val, psi_boundary):
            #     yield psi_val, next(surfs)

            else:
                for surf in surfs:
                    if isinstance(surf, Curve) and surf.is_closed and surf.enclose(R, Z):
                        surf.set_coordinates("r", "z")
                        yield psi_val, surf
                        break
                else:
                    logger.warning(f"Can not find surf at {psi_val}  ")

    def find_surfaces(self, psi_norm) -> typing.Generator[typing.Tuple[float, GeoObject], None, None]:
        psi = psi_norm * (self.psi_boundary - self.psi_axis) + self.psi_axis
        for p, surf in self.find_surfaces_by_psi(psi):
            yield (p - self.psi_axis) / (self.psi_boundary - self.psi_axis), surf

    @dataclass
    class ShapeProperty:
        psi: float | np.ndarray
        Rmin: float | np.ndarray
        Zmin: float | np.ndarray
        Rmax: float | np.ndarray
        Zmax: float | np.ndarray
        Rzmin: float | np.ndarray
        Rzmax: float | np.ndarray
        r_inboard: float | np.ndarray
        r_outboard: float | np.ndarray

    def shape_property(self, psi_norm: typing.Union[float, typing.Sequence[float]] = None) -> ShapeProperty:
        def shape_box(s: GeoObject):
            if isinstance(s, Point):
                r, z = s.points
                rmin = r
                rmax = r
                zmin = z
                zmax = z
                r_inboard = r
                r_outboard = r
                rzmin = r
                rzmax = r
            elif isinstance(s, GeoObject):
                r, z = s.points
                (rmin, zmin) = s.bbox.origin
                (rmax, zmax) = s.bbox.origin + s.bbox.dimensions
                rzmin = r[np.argmin(z)]
                rzmax = r[np.argmax(z)]
                r_inboard = s.coordinates(0.5)[0]
                r_outboard = s.coordinates(0)[0]
            else:
                raise TypeError(f"Invalid type {type(s)}")
            return rmin, zmin, rmax, zmax, rzmin, rzmax, r_inboard, r_outboard

        if psi_norm is None:
            psi_norm = self.psi_norm
        elif not isinstance(psi_norm, (np.ndarray, collections.abc.MutableSequence)):
            psi_norm = [psi_norm]

        sbox = np.asarray([[p, *shape_box(s)] for p, s in self.find_surfaces(psi_norm)], dtype=float)

        if sbox.shape[0] == 1:
            psi_norm, rmin, zmin, rmax, zmax, rzmin, rzmax, r_inboard, r_outboard = sbox[0]
        else:
            psi_norm, rmin, zmin, rmax, zmax, rzmin, rzmax, r_inboard, r_outboard = sbox.T
        if np.isscalar(psi_norm):
            return FyEquilibriumCoordinateSystem.ShapeProperty(
                psi_norm, rmin, zmin, rmax, zmax, rzmin, rzmax, r_inboard, r_outboard
            )
        else:
            return FyEquilibriumCoordinateSystem.ShapeProperty(
                psi_norm,
                Function(psi_norm, rmin, name="rmin"),
                Function(psi_norm, zmin, name="zmin"),
                Function(psi_norm, rmax, name="rmax"),
                Function(psi_norm, zmax, name="zmax"),
                Function(psi_norm, rzmin, name="rzmin"),
                Function(psi_norm, rzmax, name="rzmax"),
                Function(psi_norm, r_inboard, name="r_inboard"),
                Function(psi_norm, r_outboard, name="r_outboard"),
            )

    def _surface_integral(
        self, func: Expression, psi_norm: array_type | float = None
    ) -> typing.Tuple[ArrayLike, ArrayLike]:
        r"""
        $ V^{\prime} =  2 \pi  \int{ R / \left|\nabla \psi \right| * dl }$
        $ V^{\prime}(psi)= 2 \pi  \int{ dl * R / \left|\nabla \psi \right|}$
        """

        # r0, z0 = self.magnetic_axis

        if psi_norm is None:
            surfs_list = zip(self.grid.dims[0], self.grid.geometry)
        else:
            if isinstance(psi_norm, scalar_type):
                psi_norm = [psi_norm]
            surfs_list = self.find_surfaces(psi_norm)

        psi_norm: list = []
        res = []
        for p, surf in surfs_list:
            if isinstance(surf, Curve):
                v = surf.integral(func / self.Bpol)
            elif isinstance(surf, Point):  # o-point
                # R, Z = surf.points
                # v = (func(R, Z) if callable(func) else func) * self.ddpsi(R, Z)
                # logger.debug((R, Z, v))
                # # v *= r**2/self.ddpsi(r, z)
                # logger.warning(f"Found point pos={surf.points}  psi={p}")
                v = 0  # np.nan
            else:
                continue
                logger.warning(f"Found an island at psi={p} pos={surf}")
            res.append(v)
            psi_norm.append(p)

        if len(psi_norm) > 1:
            return np.asarray(psi_norm, dtype=float), np.asarray(res, dtype=float)
        else:
            return psi_norm[0], res[0]

    def surface_integral(self, func: Expression, psi_norm: NumericType = None) -> Expression | float:
        psi_norm, value = self._surface_integral(func, psi_norm)

        if np.isscalar(psi_norm):
            return value
        else:
            label = getattr(func, "__label__", func)
            return Function(
                psi_norm,
                value,
                name=f"surface_integral({label})",
                label=rf"<{func.__repr__()}>",
            )

    def surface_average(self, func: Expression, *xargs) -> Expression | ArrayLike:
        r"""
        $\left\langle \alpha\right\rangle \equiv\frac{2\pi}{V^{\prime}}\oint\alpha\frac{Rdl}{\left|\nabla\psi\right|}$
        """
        return self.surface_integral(func, *xargs) / self.dvolume_dpsi(*xargs)


@sp_tree(mesh="grid")
class FyEquilibriumProfiles2D(Equilibrium.TimeSlice.Profiles2D):
    _coord: FyEquilibriumCoordinateSystem = sp_property(alias="../coordinate_system")
    _profiles_1d: Equilibrium.TimeSlice.Profiles1D = sp_property(alias="../profiles_1d")
    _global_quantities: Equilibrium.TimeSlice.GlobalQuantities = sp_property(alias="../global_quantities")

    psi: Field

    j_parallel: Field

    @sp_property
    def grid(self) -> Mesh:
        g = self.get_cache("grid")
        dim1 = self.get_cache("grid/dim1")
        dim2 = self.get_cache("grid/dim2")
        mesh_type = self.get_cache("grid_type/name")
        return Mesh(dim1, dim2, type=mesh_type)

    @sp_property
    def r(self) -> Field:
        return self.grid.points[0]

    @sp_property
    def z(self) -> Field:
        return self.grid.points[1]

    @sp_property
    def psi_norm(self) -> Expression:
        """normalized psirz"""
        return (self.psi - self._coord.psi_axis) / (self._coord.psi_boundary - self._coord.psi_axis)

    @sp_property
    def phi(self) -> Expression:
        return self._profiles_1d.phi(self.psi_norm)

    @sp_property
    def theta(self) -> Expression:
        return self._coord.theta

    @sp_property
    def j_tor(self) -> Expression:
        return _R * self._profiles_1d.dpressure_dpsi(self.psi_norm) + self._profiles_1d.f_df_dpsi(self.psi_norm) / (
            _R * scipy.constants.mu_0
        )

    @sp_property(label="B_{r}")
    def b_field_r(self) -> Expression:
        """COCOS Eq.19 [O. Sauter and S.Yu. Medvedev, Computer Physics Communications 184 (2013) 293]"""
        return self.psi.pd(0, 1) / _R * (self._coord._s_RpZ * self._coord._s_Bp / self._coord._s_eBp_2PI)

    @sp_property(label="B_{z}")
    def b_field_z(self) -> Expression:
        return -self.psi.pd(1, 0) / _R * (self._coord._s_RpZ * self._coord._s_Bp / self._coord._s_eBp_2PI)

    @sp_property(label="B_{tor}")
    def b_field_tor(self) -> Expression:
        return self._profiles_1d.f(self.psi_norm) / _R

    @sp_property
    def Bpol2(self) -> Expression:
        r"""$B_{pol}= \left|\nabla \psi \right|/2 \pi R $"""
        return self.b_field_r**2 + self.b_field_z**2

    @sp_property(label="B^2")
    def B2(self) -> Expression:
        return self.b_field_r**2 + self.b_field_z**2 + self.b_field_tor**2

    @sp_property
    def grad_psi2(self) -> Expression:
        return self.psi.pd(1, 0) ** 2 + self.psi.pd(0, 1) ** 2

    @sp_property
    def grad_psi(self) -> Expression:
        return np.sqrt(self.grad_psi2)

    @sp_property
    def ddpsi(self) -> Expression:
        return np.sqrt(self.psi.pd(2, 0) * self.psi.pd(0, 2) + self.psi.pd(1, 1) ** 2)


@sp_tree
class FyEquilibriumProfiles1D(Equilibrium.TimeSlice.Profiles1D):
    _root: Equilibrium.TimeSlice = sp_property(alias="../")

    _profiles_2d: FyEquilibriumProfiles2D = sp_property(alias="../profiles_2d")

    _coord: FyEquilibriumCoordinateSystem = sp_property(alias="../coordinate_system")

    psi_norm: array_type

    f_df_dpsi: Expression

    ffprime: Expression = sp_property(alias="f_df_dpsi")

    dpressure_dpsi: Expression

    pprime: Expression = sp_property(alias="dpressure_dpsi")

    @sp_property
    def grid(self) -> CoreRadialGrid:
        psi_norm = self.psi_norm
        rho_tor_norm = self.rho_tor_norm(self.psi_norm)
        if rho_tor_norm[0] < 0:
            rho_tor_norm[0] = 0.0
        return CoreRadialGrid(
            {
                "psi_norm": psi_norm,
                "psi_axis": self._coord.psi_axis,
                "psi_boundary": self._coord.psi_boundary,
                "rho_tor_norm": rho_tor_norm,
                "rho_tor_boundary": np.sqrt(
                    np.abs(self.phi(self.psi_norm[-1]) / (scipy.constants.pi * self._coord.b0))
                ),
            }
        )

    @sp_property(label=r"\psi")
    def psi(self) -> Expression:
        return self.psi_norm * (self._coord.psi_boundary - self._coord.psi_axis) + self._coord.psi_axis

    @sp_property(label="f")
    def f(self) -> Expression:
        return np.sqrt(
            2.0 * (self._coord.psi_boundary - self._coord.psi_axis) * self.f_df_dpsi.I
            + (self._coord.b0 * self._coord.r0) ** 2
        )

    @sp_property(label=r"\phi")
    def phi(self) -> Expression:
        return self.dphi_dpsi.I * (self._coord.psi_boundary - self._coord.psi_axis)

    @sp_property(label=r"\rho_{tor}")
    def rho_tor(self) -> Expression:
        return np.sqrt(np.abs(self.phi / (scipy.constants.pi * self._coord.b0)))

    @sp_property(label=r"\bar{\rho}_{tor}")
    def rho_tor_norm(self) -> Expression:
        phi = np.asarray(self.phi)

        if np.isclose(self.psi_norm[-1], 1.0):
            phi_boundary = phi[-1]
        else:
            phi_boundary = self.phi(1.0)
        r_ = np.sqrt(phi / phi_boundary)
        if r_[0] < 0:
            r_[0] = 0.0
        return r_

    @sp_property
    def dvolume_dpsi(self) -> Expression:
        return Function(*self._coord._surface_integral(1.0), name="dvolume_dpsi", label=r"\frac{d V}{d\psi}")

    @sp_property(label=r"q")
    def q(self) -> Expression:
        return self.dphi_dpsi * (self._coord._s_eBp_2PI / (2.0 * scipy.constants.pi))

    @sp_property
    def magnetic_shear(self) -> Expression:
        return self.rho_tor * self.q.d / self.q * self.dpsi_drho_tor

    @sp_property(label=r"\frac{d\phi}{d\psi}")
    def dphi_dpsi(self) -> Expression:
        return self.f * self._coord.surface_integral(1.0 / (_R**2))

    @sp_property
    def drho_tor_dpsi(self) -> Expression:
        r"""
        $\frac{d\rho_{tor}}{d\psi} 
            =\frac{d}{d\psi}\sqrt{\frac{\Phi_{tor}}{\pi B_{0}}} \
            =\frac{1}{2\sqrt{\pi B_{0}\Phi_{tor}}}\frac{d\Phi_{tor}}{d\psi} \
            =\frac{q}{2\pi B_{0}\rho_{tor}}
        $
        """
        return 1.0 / self.dpsi_drho_tor

    @sp_property
    def dpsi_drho_tor(self) -> Expression:
        return np.abs(self._coord.b0) * self.rho_tor / self.q

    @sp_property
    def dpsi_drho_tor_norm(self) -> Expression:
        return self.dpsi_drho_tor * self._coord.rho_tor_boundary

    @sp_property
    def dpsi_norm_drho_tor_norm(self) -> Expression:
        return self.dpsi_drho_tor * self._coord.rho_tor_boundary / (self._coord.psi_boundary - self._coord.psi_axis)

    @sp_property
    def dvolume_dpsi(self) -> Expression:
        return self._coord.dvolume_dpsi

    @sp_property(label=r"\frac{dV}{d\rho_{tor}}")
    def dvolume_drho_tor(self) -> Expression:
        return self._coord._s_eBp_2PI * np.abs(self._coord.b0) * self.dvolume_dpsi * self.dpsi_drho_tor

    @sp_property
    def volume(self) -> Expression:
        return self.dvolume_dpsi.I * (self.grid.psi_boundary - self.grid.psi_axis)

    @sp_property
    def area(self) -> Expression:
        return self.darea_dpsi.I * (self.grid.psi_boundary - self.grid.psi_axis)

    @sp_property
    def darea_dpsi(self) -> Expression:
        """FIXME: just a simple approximation!"""
        return self.dvolume_dpsi / ((2.0 * scipy.constants.pi) * self._coord.r0)

    @sp_property
    def darea_drho_tor(self) -> Expression:
        return self.darea_dpsi * self.dpsi_drho_tor

    @sp_property
    def surface(self) -> Expression:
        return self.dvolume_drho_tor * self.gm7

    @sp_property
    def dphi_dvolume(self) -> Expression:
        return self.f * self.gm1

    @sp_property
    def gm1(self) -> Expression:
        return self._coord.surface_average(1.0 / (_R**2))

    @sp_property
    def gm2(self) -> Expression:
        return self._coord.surface_average(self._profiles_2d.grad_psi2 / (_R**2)) / (self.dpsi_drho_tor**2)

    @sp_property
    def gm3(self) -> Expression:
        return self._coord.surface_average(self._profiles_2d.grad_psi2) / (self.dpsi_drho_tor**2)

    @sp_property
    def gm4(self) -> Expression:
        return self._coord.surface_average(1.0 / self._profiles_2d.B2)

    @sp_property
    def gm5(self) -> Expression:
        return self._coord.surface_average(self._profiles_2d.B2)

    @sp_property
    def gm6(self) -> Expression:
        return self._coord.surface_average(self._profiles_2d.grad_psi2 / self._profiles_2d.B2) / (self.dpsi_drho_tor**2)

    @sp_property
    def gm7(self) -> Expression:
        return self._coord.surface_average(np.sqrt(self._profiles_2d.grad_psi2)) / self.dpsi_drho_tor

    @sp_property
    def gm8(self) -> Expression:
        return self._coord.surface_average(_R)

    @sp_property
    def gm9(self) -> Expression:
        return self._coord.surface_average(1.0 / _R)

    # 描述磁面形状
    @functools.cached_property
    def _shape_property(self) -> FyEquilibriumCoordinateSystem.ShapeProperty:
        return self._coord.shape_property(self.psi_norm)

    @sp_property
    def minor_radius(self) -> Expression:
        return (self._shape_property.Rmax - self._shape_property.Rmin) * 0.5

    @sp_property
    def major_radius(self) -> Expression:
        return (self._shape_property.Rmax + self._shape_property.Rmin) * 0.5

    @sp_property
    def magnetic_z(self) -> Expression:
        return (self._shape_property.Zmax + self._shape_property.Zmin) * 0.5

    @sp_property
    def r_inboard(self) -> Expression:
        return self._shape_property.r_inboard

    @sp_property
    def r_outboard(self) -> Expression:
        return self._shape_property.r_outboard

    @sp_property
    def elongation(self) -> Expression:
        return (self._shape_property.Zmax - self._shape_property.Zmin) / (
            self._shape_property.Rmax - self._shape_property.Rmin
        )

    @sp_property
    def elongation_upper(self) -> Expression:
        return (self._shape_property.Zmax - (self._shape_property.Zmax + self._shape_property.Zmin) * 0.5) / (
            self._shape_property.Rmax - self._shape_property.Rmin
        )

    @sp_property
    def elongation_lower(self) -> Expression:
        return ((self._shape_property.Zmax + self._shape_property.Zmin) * 0.5 - self._shape_property.Zmin) / (
            self._shape_property.Rmax - self._shape_property.Rmin
        )

    @sp_property
    def triangularity_upper(self) -> Expression:
        return (
            ((self._shape_property.Rmax - self._shape_property.Rmin) * 0.5 - self._shape_property.Rzmax)
            / (self._shape_property.Rmax - self._shape_property.Rmin)
            * 2
        )

    @sp_property
    def triangularity_lower(self) -> Expression:
        return (
            ((self._shape_property.Rmax + self._shape_property.Rmin) * 0.5 - self._shape_property.Rzmin)
            / (self._shape_property.Rmax - self._shape_property.Rmin)
            * 2
        )

    @sp_property
    def triangularity(self) -> Expression:
        psi = self.grid.psi[1:]
        res = (
            (self._shape_property.Rzmax(psi) - self._shape_property.Rzmin(psi))
            / (self._shape_property.Rmax(psi) - self._shape_property.Rmin(psi))
            * 2
        )
        return Function(psi, res, name="triangularity")

    @sp_property
    def squareness(self) -> Expression:
        return 0.0

    @sp_property
    def trapped_fraction(self) -> Expression:
        """Trapped particle fraction[-]
        Tokamak 3ed, 14.10
        """
        epsilon = self.rho_tor(self.psi) / self._coord._R0
        return 1.0 - (1 - epsilon) ** 2 / np.sqrt(1.0 - epsilon**2) / (1 + 1.46 * np.sqrt(epsilon))

    @sp_property
    def plasma_current(self) -> Expression:
        return self.gm2 * self.dvolume_drho_tor / self.dpsi_drho_tor / scipy.constants.mu_0

    @sp_property
    def j_tor(self) -> Expression:
        return self.plasma_current.d / self.dvolume_dpsi * self._coord.r0

    @sp_property
    def j_parallel(self) -> Expression:
        return self._coord.surface_average(dot(self._coord.j, self._coord.B) / np.sqrt(self._coord.B2))

        # fvac = self._coord._fvac
        # d = np.asarray(function_like(np.asarray(self.volume),
        #                              np.asarray(fvac*self.plasma_current/self.f)).pd())
        # return self._coord._R0*(self.f / fvac)**2 * d


class FyEquilibriumGlobalQuantities(Equilibrium.TimeSlice.GlobalQuantities):
    _root: Equilibrium.TimeSlice = sp_property(alias="../")

    _coord: FyEquilibriumCoordinateSystem = sp_property(alias="../coordinate_system")

    psi_axis: float

    psi_boundary: float

    magnetic_axis: Point = sp_property(alias="../coordinate_system/magnetic_axis")

    @sp_property
    def b_field_tor_axis(self) -> float:
        return (self._root.profiles_2d.b_field_tor(self.magnetic_axis.r, self.magnetic_axis.z),)

    @sp_property
    def q_axis(self) -> float:
        return self._root.profiles_1d.q(0.0)

    @sp_property
    def q_95(self) -> float:
        return self._root.profiles_1d.q(0.95)

    @sp_property
    def q_min(self) -> typing.Tuple[float, float]:
        q = np.asarray(self._root.profiles_1d.q)
        idx = np.argmin(q)
        return q[idx], self._root.profiles_1d.rho_tor_norm[idx]


class FyEquilibriumBoundary(Equilibrium.TimeSlice.Boundary):
    _coord: FyEquilibriumCoordinateSystem = sp_property(alias="../coordinate_system")

    # psi_norm: float

    psi: float

    phi: float

    rho_tor_norm: float

    @sp_property(coordinates="r z")
    def outline(self) -> Curve:
        _, surf = next(self._coord.find_surfaces(self.psi_norm))
        return surf

    @functools.cached_property
    def _shape_property(self) -> FyEquilibriumCoordinateSystem.ShapeProperty:
        return self._coord.shape_property(self.psi_norm)

    @sp_property
    def geometric_axis(self) -> Point:
        return {
            "r": (self._shape_property.Rmin + self._shape_property.Rmax) * 0.5,
            "z": (self._shape_property.Zmin + self._shape_property.Zmax) * 0.5,
        }

    @sp_property
    def minor_radius(self) -> float:
        return (self._shape_property.Rmax - self._shape_property.Rmin) * 0.5

    @sp_property
    def elongation(self) -> float:
        return (self._shape_property.Zmax - self._shape_property.Zmin) / (
            self._shape_property.Rmax - self._shape_property.Rmin
        )

    @sp_property
    def elongation_upper(self) -> float:
        return (self._shape_property.Zmax - (self._shape_property.Zmax + self._shape_property.Zmin) * 0.5) / (
            self._shape_property.Rmax - self._shape_property.Rmin
        )

    @sp_property
    def elongation_lower(self) -> float:
        return ((self._shape_property.Zmax + self._shape_property.Zmin) * 0.5 - self._shape_property.Zmin) / (
            self._shape_property.Rmax - self._shape_property.Rmin
        )

    @sp_property(coordinate1="../psi")
    def triangularity(self) -> float:
        return (
            (self._shape_property.Rzmax - self._shape_property.Rzmin)
            / (self._shape_property.Rmax - self._shape_property.Rmin)
            * 2
        )

    @sp_property
    def triangularity_upper(self) -> float:
        return (
            ((self._shape_property.Rmax + self._shape_property.Rmin) * 0.5 - self._shape_property.Rzmax)
            / (self._shape_property.Rmax - self._shape_property.Rmin)
            * 2
        )

    @sp_property
    def triangularity_lower(self) -> float:
        return (
            ((self._shape_property.Rmax + self._shape_property.Rmin) * 0.5 - self._shape_property.Rzmin)
            / (self._shape_property.Rmax - self._shape_property.Rmin)
            * 2
        )

    @sp_property
    def x_point(self) -> List[Point]:
        return [Point(p.r, p.z) for p in self._coord.x_point]

    @sp_property
    def strike_point(self) -> List[Point]:
        return

    @sp_property
    def active_limiter_point(self) -> List[Point]:
        return NotImplemented


@sp_tree
class FyEquilibriumBoundarySeparatrix(FyEquilibriumBoundary):
    _coord: FyEquilibriumCoordinateSystem = sp_property(alias="../coordinate_system")

    @sp_property(coordinates="r z")
    def outline(self) -> GeoObjectSet:
        """RZ outline of the plasma boundary"""
        return GeoObjectSet([surf for _, surf in self._coord.find_surfaces(self.psi_norm)])

    psi_norm: float = 1.0


@sp_tree
class FyEquilibriumTimeSlice(Equilibrium.TimeSlice):
    vacuum_toroidal_field: VacuumToroidalField

    global_quantities: FyEquilibriumGlobalQuantities

    profiles_1d: FyEquilibriumProfiles1D

    profiles_2d: FyEquilibriumProfiles2D

    boundary: FyEquilibriumBoundary

    boundary_separatrix: FyEquilibriumBoundarySeparatrix

    coordinate_system: FyEquilibriumCoordinateSystem = sp_property(default_value={})


@sp_tree
class FyEqAnalyze(Equilibrium):
    """
    Magnetic surface analyze 磁面分析工具
    =============================
    input:
        - vacuum_toroidal_field.b0, vacuum_toroidal_field.r0
        - f, Diamagnetic function (F=R B_Phi)
        - profiles_2d.psi (RZ 2D)

    output：
        - 识别 O,X point
        - 识别 Separatrix, boundary
        - Surface average

    """

    code: Code = {"name": "fy_eq", "copyright": "FUYUN"}

    TimeSlice = FyEquilibriumTimeSlice

    time_slice: TimeSeriesAoS[FyEquilibriumTimeSlice]


Equilibrium.register("fy_eq", FyEqAnalyze)



import collections

import numpy as np
from fytok.modules.core_profiles import CoreProfiles
from fytok.modules.core_transport import CoreTransport
from fytok.modules.equilibrium import Equilibrium
from spdm.utils.logger import logger
from spdm.utils.typing import array_type
from spdm.utils.tags import _not_found_
from spdm.core.sp_property import sp_tree


@sp_tree
class PredefinedTransport(CoreTransport.Model):
    """ """

    identifier = "predefined"
    code = {"name": "predefined", "description": f"predefined", "copyright": "fytok"}

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D, *args, **kwargs) -> CoreTransport.Model.TimeSlice:
        current: CoreTransport.Model.TimeSlice = super().fetch(*args, **kwargs)

        rho_tor_norm = profiles_1d.rho_tor_norm

        eq: Equilibrium.TimeSlice = self.inports["equilibrium/time_slice/current"].fetch()

        eq_1d = eq.profiles_1d

        B0 = np.abs(eq.vacuum_toroidal_field.b0)

        R0 = eq.vacuum_toroidal_field.r0

        # rho_tor_norm = Variable(0, name="rho_tor_norm", label=r"\bar{\rho}_{tor}")

        _x = rho_tor_norm

        # Core profiles
        r_ped = 0.96  # np.sqrt(0.88)

        # Core Transport
        Cped = 0.17
        Ccore = 0.4

        delta = np.heaviside(_x - r_ped, 0.5)

        chi = (Ccore * (1.0 + 3 * (_x**2))) * (1 - delta) + Cped * delta
        chi_e = Ccore * (1.0 + 3 * (_x**2)) * (1 - delta) * 0.5 + Cped * delta

        D = 0.1 * (chi + chi_e)

        v_pinch_ne = 0.6 * D * _x / R0
        v_pinch_Te = 2.5 * chi_e * _x / R0
        v_pinch_ni = D * _x / R0
        v_pinch_Ti = chi * _x / R0

        current.flux_multiplier = 3 / 2

        current.profiles_1d.grid_d = eq_1d.grid
        current.profiles_1d.electrons.particles.d = D
        current.profiles_1d.electrons.particles.v = -v_pinch_ne
        current.profiles_1d.electrons.energy.d = chi_e
        current.profiles_1d.electrons.energy.v = -v_pinch_Te

        current.profiles_1d.ion.extend(
            [
                {"@name": "D", "particles": {"d": D, "v": -v_pinch_ni}, "energy": {"d": chi, "v": -v_pinch_Ti}},
                {"@name": "T", "particles": {"d": D, "v": -v_pinch_ni}, "energy": {"d": chi, "v": -v_pinch_Ti}},
                {"@name": "He", "particles": {"d": D, "v": -v_pinch_ni}, "energy": {"d": chi, "v": -v_pinch_Ti}},
            ]
        )

        return current


CoreTransport.Model.register(["predefined"], PredefinedTransport)



from __future__ import annotations
import collections
import functools
import typing
from dataclasses import dataclass
from enum import IntFlag
import numpy as np

from spdm.core.path import Path
from spdm.core.actor import Actor
from spdm.core.aos import AoS
from spdm.core.field import Field
from spdm.core.expression import Expression, zero
from spdm.core.function import Function
from spdm.core.htree import Dict, HTree, List
from spdm.core.signal import Signal, SignalND
from spdm.core.sp_property import SpTree, sp_property, sp_tree, PropertyTree
from spdm.core.time_series import TimeSeriesAoS, TimeSlice
from spdm.core.domain import PPolyDomain

from spdm.geometry.curve import Curve
from spdm.utils.typing import array_type, is_array, as_array
from spdm.utils.tags import _not_found_
from spdm.view import sp_view as sp_view

from ..utils.logger import logger
from ..utils.envs import FY_JOBID


@sp_tree
class IDSProperties:
    comment: str
    homogeneous_time: int
    provider: str
    creation_date: str
    version_put: SpTree
    provenance: SpTree


@sp_tree
class Library:
    name: str
    commit: str
    version: str
    repository: str
    parameters: SpTree


@sp_tree
class Code:
    name: str = "default"
    """代码名称，也是调用 plugin 的 identifier"""

    module_path: str
    """模块路径， 可用于 import 模块"""

    commit: str
    version: str = "0.0.0"
    copyright: str = "NO_COPYRIGHT_STATEMENT"
    repository: str = ""
    output_flag: array_type
    library: List[Library]
    parameters: PropertyTree = {}
    """指定参数列表，代码调用时所需，但不在由 Module 定义的参数列表中的参数。 """

    def __str__(self) -> str:
        return "-".join([s for s in [self.name, self.version.replace(".", "_")] if isinstance(s, str)])

    def __repr__(self) -> str:
        desc = {
            "name": self.name,
            "version": self.version,
            "copyright": self.copyright,
        }

        return ", ".join(
            [
                f"{key}='{value}'"
                for key, value in desc.items()
                if value is not _not_found_ and value is not None and value != ""
            ]
        )


@sp_tree
class Identifier:
    def __init__(self, *args, **kwargs):
        if len(args) == 0:
            pass
        elif isinstance(args[0], str):
            args[0] = {"name": args[0]}
        elif isinstance(args[0], int):
            args[0] = {"int": args[0]}
        super().__init__(*args, **kwargs)

    name: str
    index: int
    description: str


@sp_tree
class Module(Actor):

    def __new__(cls, *args, **kwargs) -> typing.Type[typing.Self]:
        pth = Path("code/name")
        plugin_name = None
        if len(args) > 0 and isinstance(args[0], dict):
            plugin_name = pth.get(args[0], None)
        if plugin_name is None:
            plugin_name = pth.get(kwargs, None)
        if plugin_name is None:
            plugin_name = Path("default_value/name").get(cls.code, None)
        if plugin_name == "default":
            plugin_name = None
        return super().__new__(cls, {"$class": plugin_name})

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.code.name:
            self.code.name = self.__class__.__name__

        self.code.module_path = self.__module__ + "." + self.__class__.__name__

        logger.verbose(f"Initialize module {self.code} ")

    code: Code = {}
    """ 对于 Module 的一般性说明。 
        @note code 在 __init__ 时由初始化参数定义。"""

    @property
    def tag(self) -> str:
        return f"{FY_JOBID}/{self.code.module_path}"

    def refresh(self, *args, **kwargs) -> typing.Type[TimeSlice]:
        """更新当前 Actor 的状态。
        更新当前状态树 （time_slice），并执行 self.iteration+=1

        """
        logger.verbose(f"Refresh module {self.code.module_path}")

        current = super().refresh(*args, **kwargs)

        return current


class IDS(Module):
    """Base class of IDS"""

    ids_properties: IDSProperties

    """Interface Data Structure properties. This element identifies the node above as an IDS"""

    def _repr_svg_(self) -> str:
        if hasattr(self.__class__, "__view__"):
            try:
                res = sp_view.display(self.__view__(), output="svg")
            except Exception as error:
                raise RuntimeError(f"{self}") from error
        else:
            res = None
        return res


@sp_tree
class RZTuple:
    r: typing.Any
    z: typing.Any


@sp_tree
class PointRZ:  # utilities._T_rz0d_dynamic_aos
    r: float
    z: float


@sp_tree
class CurveRZ:  # utilities._T_rz1d_dynamic_aos
    r: array_type
    z: array_type


@sp_tree
class VacuumToroidalField:
    r0: float
    b0: float


@sp_tree
class CoreRadialGrid:
    def __init__(self, *args, **kwargs) -> None:
        SpTree.__init__(self, *args, **kwargs)
        # PPolyDomain.__init__(self, self._cache["psi_norm"])
        # assert isinstance(self.psi_axis, float), f"psi_axis must be specified  {self.psi_axis}"
        # assert isinstance(self.psi_boundary, float), f"psi_boundary must be specified {self.psi_boundary}"
        # assert isinstance(self.rho_tor_boundary, float), f"rho_tor_boundary must be specified {self.rho_tor_boundary}"
        # assert self.rho_tor_norm[0] >= 0 and self.rho_tor_norm[-1] <= 1.0, f"illegal rho_tor_norm {self.rho_tor_norm}"
        # assert self.psi_norm[0] >= 0 and self.psi_norm[-1] <= 1.0, f"illegal psi_norm {self.psi_norm}"

    def __copy__(self) -> CoreRadialGrid:
        return CoreRadialGrid(
            {
                "psi_norm": self.psi_norm,
                "rho_tor_norm": self.rho_tor_norm,
                "psi_axis": self.psi_axis,
                "psi_boundary": self.psi_boundary,
                "rho_tor_boundary": self.rho_tor_boundary,
            }
        )

    def __serialize__(self, dumper=None):
        return HTree._do_serialize(
            {
                "psi_norm": self.psi_norm,
                "rho_tor_norm": self.rho_tor_norm,
                "psi_axis": self.psi_axis,
                "psi_boundary": self.psi_boundary,
                "rho_tor_boundary": self.rho_tor_boundary,
            },
            dumper,
        )

    def remesh(self, rho_tor_norm=None, *args, psi_norm=None, **kwargs) -> CoreRadialGrid:
        """Duplicate the grid with new rho_tor_norm or psi_norm"""

        if isinstance(rho_tor_norm, array_type):
            psi_norm = Function(self.rho_tor_norm, self.psi_norm)(rho_tor_norm)
            if psi_norm[0] < 0:
                psi_norm[0] = 0.0
        elif isinstance(psi_norm, array_type):
            rho_tor_norm = Function(self.psi_norm, self.rho_tor_norm)(psi_norm)
            if rho_tor_norm[0] < 0:
                psi_norm[0] = 0.0

        else:
            rho_tor_norm = self.rho_tor_norm
            psi_norm = self.psi_norm
        # if rho_tor_norm is None or rho_tor_norm is _not_found_:
        #     if psi_norm is _not_found_ or psi_norm is None:
        #         psi_norm = self.psi_norm
        #         rho_tor_norm = self.rho_tor_norm
        #     else:
        #         rho_tor_norm = Function(
        #             self.psi_norm,
        #             self.rho_tor_norm,
        #             name="rho_tor_norm",
        #             label=r"\bar{\rho}",
        #         )(psi_norm)
        # else:
        #     rho_tor_norm = np.asarray(rho_tor_norm)

        # if psi_norm is _not_found_ or psi_norm is None:
        #     psi_norm = Function(
        #         self.rho_tor_norm,
        #         self.psi_norm,
        #         name="psi_norm",
        #         label=r"\bar{\psi}",
        #     )(rho_tor_norm)

        # else:
        #     psi_norm = np.asarray(psi_norm)

        return CoreRadialGrid(
            {
                "rho_tor_norm": rho_tor_norm,
                "psi_norm": psi_norm,
                "psi_axis": self.psi_axis,
                "psi_boundary": self.psi_boundary,
                "rho_tor_boundary": self.rho_tor_boundary,
            }
        )

    def fetch(self, first=None, *args, psi_norm=None, **kwargs) -> CoreRadialGrid:
        if isinstance(first, array_type):
            rho_tor_norm = first
        else:
            rho_tor_norm = getattr(first, "rho_tor_norm", kwargs.pop("rho_tor_norm", None))

        if psi_norm is None and isinstance(first, SpTree):
            psi_norm = getattr(first, "psi_norm", None)

        return self.remesh(rho_tor_norm, *args, psi_norm=psi_norm, **kwargs)

    psi_axis: float
    psi_boundary: float
    psi_norm: array_type

    phi_boundary: float
    phi_norm: array_type

    rho_tor_boundary: float
    rho_tor_norm: array_type

    @sp_property
    def psi(self) -> array_type:
        return self.psi_norm * (self.psi_boundary - self.psi_axis) + self.psi_axis

    @sp_property
    def phi(self) -> array_type:
        return self.phi_norm * self.phi_boundary

    @sp_property
    def rho_tor(self) -> array_type:
        return self.rho_tor_norm * self.rho_tor_boundary

    @sp_property
    def rho_pol_norm(self) -> array_type:
        return np.sqrt(self.psi_norm)


@sp_tree
class CoreVectorComponents:
    """Vector components in predefined directions"""

    radial: Expression
    """ Radial component"""

    diamagnetic: Expression
    """ Diamagnetic component"""

    parallel: Expression
    """ Parallel component"""

    poloidal: Expression
    """ Poloidal component"""

    toroidal: Expression
    """ Toroidal component"""


class DetectorAperture:  # (utilities._T_detector_aperture):
    def __view__(self, view="RZ", **kwargs):
        geo = {}
        styles = {}
        return geo, styles


@sp_tree
class PlasmaCompositionIonState:
    label: str
    z_min: float = sp_property(units="Elementary Charge Unit")
    z_max: float = sp_property(units="Elementary Charge Unit")
    electron_configuration: str
    vibrational_level: float = sp_property(units="Elementary Charge Unit")
    vibrational_mode: str


@sp_tree
class PlasmaCompositionSpecies:
    label: str
    a: float  # = sp_property(units="Atomic Mass Unit", )
    z_n: float  # = sp_property(units="Elementary Charge Unit", )


@sp_tree
class PlasmaCompositionNeutralElement(SpTree):
    a: float  # = sp_property(units="Atomic Mass Unit", )
    z_n: float  # = sp_property(units="Elementary Charge Unit", )
    atoms_n: int


@sp_tree
class PlasmaCompositionIons:
    label: str
    element: AoS[PlasmaCompositionNeutralElement]
    z_ion: float  # = sp_property( units="Elementary Charge Unit")
    state: PlasmaCompositionIonState


class PlasmaCompositionNeutralState:
    label: str
    electron_configuration: str
    vibrational_level: float  # = sp_property(units="Elementary Charge Unit")
    vibrational_mode: str
    neutral_type: str


class PlasmaCompositionNeutral:
    label: str
    element: AoS[PlasmaCompositionNeutralElement]
    state: PlasmaCompositionNeutralState


@sp_tree
class DistributionSpecies(SpTree):
    type: str
    ion: PlasmaCompositionIons
    neutral: PlasmaCompositionNeutral


# __all__ = ["IDS", "Module", "Code", "Library",
#            "DetectorAperture", "CoreRadialGrid", "PointRZ",   "CurveRZ",
#            "array_type", "Function", "Field",
#            "HTree", "List", "Dict", "SpTree", "sp_property",
#            "AoS", "TimeSeriesAoS", "TimeSlice",
#            "Signal", "SignalND", "Identifier"
#            "IntFlag"]



import typing
from spdm.core.aos import AoS
from spdm.geometry.circle import Circle
from spdm.core.geo_object import GeoObject
from spdm.geometry.polyline import Polyline
from spdm.utils.tags import _not_found_

from ..ontology import wall

from ..utils.logger import logger


class Wall(wall._T_wall):
    """Description of the torus wall and its interaction with the plasma"""

    Description2D = wall._T_wall_2d

    description_2d: AoS[Description2D]

    def __view__(self, view_point="RZ", **kwargs) -> GeoObject | typing.List[GeoObject]:
        geo = {}

        desc = self.description_2d[0]  # 0 for equilibrium codes

        match view_point.lower():
            case "top":
                vessel_r = desc.vessel.unit[0].annular.outline_outer.r
                # vessel_z = desc.vessel.unit[0].annular.outline_outer.z
                geo["vessel_outer"] = [Circle(0.0, 0.0, vessel_r.min()), Circle(0.0, 0.0, vessel_r.max())]

            case "rz":
                if desc.limiter.unit[0].outline.r is not _not_found_:
                    geo["limiter"] = Polyline(
                        desc.limiter.unit[0].outline.r,
                        desc.limiter.unit[0].outline.z,
                        styles={"$matplotlib": {"edgecolor": "green"}},
                    )
                else:
                    units = []
                    for unit in desc.vessel.unit:
                        if unit.annular is not _not_found_:
                            units.append(
                                {
                                    "annular": {
                                        "vessel_inner": Polyline(
                                            unit.annular.outline_inner.r,
                                            unit.annular.outline_inner.z,
                                            styles={"$matplotlib": {"edgecolor": "blue"}},
                                        ),
                                        "vessel_outer": Polyline(
                                            unit.annular.outline_outer.r,
                                            unit.annular.outline_outer.z,
                                            styles={"$matplotlib": {"edgecolor": "blue"}},
                                        ),
                                    }
                                }
                            )

                        else:
                            elements = []
                            for element in unit.element:
                                elements.append(Polyline(element.outline.r, element.outline.z, name=element.name))
                            units.append({"element": elements})

                        geo["unit"] = units

        return geo



from spdm.core.geo_object import GeoObject
from spdm.geometry.polygon import Rectangle
from ..utils.logger import logger
from ..ontology import pf_active


class PFActive(pf_active._T_pf_active):
    def __view__(self, view_point="RZ", **kwargs) -> GeoObject:
        geo = {}

        match view_point.lower():
            case "rz":
                geo_coils = []
                for coil in self.coil:
                    rect = coil.element[0].geometry.rectangle
                    geo_coils.append(
                        Rectangle(
                            rect.r - rect.width / 2.0,
                            rect.z - rect.height / 2.0,
                            rect.width,
                            rect.height,
                            name=coil.name,
                            styles={"$matplotlib": {"color": "black"}, "text": True},
                        )
                    )

                geo["coil"] = geo_coils

        return geo 



from __future__ import annotations
from copy import copy, deepcopy
import math

from spdm.core.aos import AoS
from spdm.core.sp_property import sp_property, sp_tree
from spdm.core.time_series import TimeSeriesAoS
from spdm.core.expression import Expression, Variable, zero, one
from spdm.utils.tags import _not_found_

from .core_profiles import CoreProfiles
from .equilibrium import Equilibrium
from .utilities import *
from ..utils.atoms import atoms

from ..ontology import core_sources


@sp_tree
class CoreSourcesSpecies(SpTree):
    """Source terms related to electrons"""

    @sp_tree
    class _Decomposed(SpTree):
        """Source terms decomposed for the particle transport equation, assuming
        core_radial_grid 3 levels above"""

        implicit_part: Expression
        """ Implicit part of the source term, i.e. to be multiplied by the equation's
        primary quantity"""

        explicit_part: Expression
        """ Explicit part of the source term"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        if self.z is _not_found_:
            ion = atoms[self.label]
            self.z = ion.z
            self.a = ion.a

    label: str = sp_property(alias="@name")
    """ String identifying the neutral species (e.g. H, D, T, He, C, ...)"""

    z: int
    """ Charge number of the neutral species"""

    a: float
    """ Mass number of the neutral species"""

    particles: Expression = sp_property(units="s^-1.m^-3", default_value=0)
    """Source term for electron density equation"""

    particles_decomposed: _Decomposed

    @sp_property(units="s^-1")
    def particles_inside(self) -> Expression:
        """Electron source inside the flux surface. Cumulative volume integral of the
        source term for the electron density equation."""
        return self.particles.I

    energy: Expression = sp_property(units="W.m^-3", default_value=0)
    """Source term for the electron energy equation"""

    energy_decomposed: _Decomposed

    @sp_property(units="W")
    def power_inside(self) -> Expression:
        """Power coupled to electrons inside the flux surface. Cumulative volume integral
        of the source term for the electron energy equation"""
        return self.energy.I

    momentum: CoreVectorComponents = sp_property(units="kg.m^-1.s^-2")


@sp_tree
class CoreSourcesElectrons(CoreSourcesSpecies):
    label: str = "electron"
    """ String identifying the neutral species (e.g. H, D, T, He, C, ...)"""


@sp_tree
class CoreSourcesNeutral(core_sources._T_core_sources_source_profiles_1d_neutral):
    pass


@sp_tree(domain="grid/rho_tor_norm")
class CoreSourcesProfiles1D(core_sources._T_core_sources_source_profiles_1d):
    grid: CoreRadialGrid
    """ Radial grid"""

    total_ion_energy: Expression = sp_property(units="W.m^-3")
    """Total ion energy source"""

    @sp_property(units="W")
    def total_ion_power_inside(self) -> Expression:
        return self.torque_tor_inside.I

    momentum_tor: Expression

    torque_tor_inside: Expression

    momentum_tor_j_cross_b_field: Expression

    j_parallel: Expression

    current_parallel_inside: Expression

    conductivity_parallel: Expression

    Electrons = CoreSourcesElectrons
    electrons: CoreSourcesElectrons

    Ion = CoreSourcesSpecies
    ion: AoS[CoreSourcesSpecies]

    Neutral = CoreSourcesNeutral
    neutral: AoS[CoreSourcesNeutral]


@sp_tree
class CoreSourcesGlobalQuantities(core_sources._T_core_sources_source_global):
    pass


@sp_tree
class CoreSourcesTimeSlice(TimeSlice):
    Profiles1D = CoreSourcesProfiles1D

    GlobalQuantities = CoreSourcesGlobalQuantities

    profiles_1d: CoreSourcesProfiles1D

    global_quantities: CoreSourcesGlobalQuantities


@sp_tree
class CoreSourcesSource(Module):
    _plugin_prefix = "fytok.modules.core_sources.source."

    identifier: str

    species: DistributionSpecies

    TimeSlice = CoreSourcesTimeSlice

    time_slice: TimeSeriesAoS[CoreSourcesTimeSlice]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def preprocess(self, *args, **kwargs) -> CoreSourcesTimeSlice:
        current: CoreSourcesTimeSlice = super().preprocess(*args, **kwargs)

        grid = current.get_cache("profiles_1d/grid", _not_found_)

        if not isinstance(grid, CoreRadialGrid):
            eq_grid: CoreRadialGrid = self.inports["/equilibrium/time_slice/0/profiles_1d/grid"].fetch()

            if isinstance(grid, dict):
                new_grid = grid
                if not isinstance(grid.get("psi_axis", _not_found_), float):
                    new_grid["psi_axis"] = eq_grid.psi_axis
                    new_grid["psi_boundary"] = eq_grid.psi_boundary
                    new_grid["rho_tor_boundary"] = eq_grid.rho_tor_boundary
                # new_grid = {
                #     **eq_grid._cache,
                #     **{k: v for k, v in grid.items() if v is not _not_found_ and v is not None},
                # }
            else:
                rho_tor_norm = kwargs.get("rho_tor_norm", _not_found_)

                if rho_tor_norm is _not_found_:
                    rho_tor_norm = self.code.parameters.rho_tor_norm

                new_grid = eq_grid.remesh(rho_tor_norm)

            current["profiles_1d/grid"] = new_grid

        return current

    def fetch(self, profiles_1d: CoreProfiles.TimeSlice.Profiles1D, *args, **kwargs) -> CoreSourcesTimeSlice:
        return super().fetch(profiles_1d.rho_tor_norm, *args, **kwargs)

    def flush(self) -> CoreSourcesTimeSlice:
        super().flush()

        current: CoreSourcesTimeSlice = self.time_slice.current

        profiles_1d: CoreProfiles.TimeSlice.Profiles1D = self.inports["core_profiles/time_slice/0/profiles_1d"].fetch()
        # eq_grid: CoreRadialGrid = self.inports["equilibrium/time_slice/0/profiles_1d/grid"].fetch()

        current.update(self.fetch(profiles_1d)._cache)

        return current

    def refresh(
        self,
        *args,
        equilibrium: Equilibrium = None,
        core_profiles: CoreProfiles = None,
        **kwargs,
    ) -> CoreSourcesTimeSlice:
        return super().refresh(*args, equilibrium=equilibrium, core_profiles=core_profiles, **kwargs)


@sp_tree
class CoreSources(IDS):
    Source = CoreSourcesSource

    source: AoS[CoreSourcesSource]

    def initialize(self, *args, **kwargs) -> None:
        super().initialize(*args, **kwargs)
        for source in self.source:
            source.initialize()

    def refresh(self, *args, equilibrium: Equilibrium = None, core_profiles: CoreProfiles = None, **kwargs):
        super().refresh(*args, **kwargs)

        kwargs.pop("time", None)

        for source in self.source:
            source.refresh(time=self.time, equilibrium=equilibrium, core_profiles=core_profiles, **kwargs)

    def advance(self, *args, equilibrium: Equilibrium = None, core_profiles: CoreProfiles = None, **kwargs):
        super().advance(*args, **kwargs)

        for source in self.source:
            source.advance(time=self.time, equilibrium=equilibrium, core_profiles=core_profiles, **kwargs)

    def flush(self):
        super().flush()

        for source in self.source:
            source.flush()




from ..ontology import magnetics
from spdm.core.geo_object import GeoObject
from spdm.geometry.point import Point
from spdm.utils.tags import _not_found_


class Magnetics(magnetics._T_magnetics):
    """Magnetic diagnostics for equilibrium identification and plasma shape control.
    """

    def __view__(self, view_point="RZ", **kwargs) -> GeoObject:
        geo = {}
        match view_point.lower():
            case "rz":
                if self.b_field_tor_probe is not _not_found_:
                    geo["b_field_tor_probe"] = [Point(p.position[0].r,  p.position[0].z, name=p.name)
                                                for p in self.b_field_tor_probe]
                if self.flux_loop is not _not_found_:
                    geo["flux_loop"] = [Point(p.position[0].r,  p.position[0].z, name=p.name) for p in self.flux_loop]

        return geo 
